""" Goniometer positions, calculations and conversions.

    Positions are normally constructed by the Position() factory
    function from this module. None of the other classes and routines
    available in this module are widely used outside this implementation.

    The Position() objects constructed by the Position() factory can
    be:

    * converted to another type using as_?() functions where ? is
      a one-letter uppercase abbreviation of the type.

    * modified using the 'With()' method.

    * queried using attribute references.
"""

__version__='$Id: position.py,v 1.6 2021/02/15 10:22:55 schreurs Exp $'
#
# (C) 1997--2004 Rob W.W. Hooft, Bruker AXS BV
#

import numpy,sys
import projtls
from misc import matrix

class error(Exception):
    """ Exception raised when a conversion can not be performed.
    """
    pass

def normalangle(a):
    """ Bring angle 'a' between -pi and +pi.
    """
    while a<=-numpy.pi:
        a=a+2.0*numpy.pi
    while a>numpy.pi:
        a=a-2.0*numpy.pi
    return a

def sameangle(a1,a2,crit=0.003*projtls.degrees):
    """ Return true if 'a1' and 'a2' are the same within 'crit'.
    """
    d=abs(a1-a2)
    while d>numpy.pi:
        d=d-2.0*numpy.pi
    d=abs(d)
    #if d>=crit:
    #print(a1/projtls.degrees,"!=",a2/projtls.degrees)
    return d<crit

def angleinrange(a,amin,amax,crit=0.001*projtls.degrees):
    """ Return true if 'amin <= a <= amax' is fulfilled Within 'crit'.
    """
    amin=amin-crit
    amax=amax+crit
    while a<=amin:
        a=a+2.0*numpy.pi
    while a>amax:
        a=a-2.0*numpy.pi
    return amin<=a<=amax

def movetorange(a,amin,amax):
    """ Bring angle 'a' between 'amin' and 'amax'. Raise a
        'ValueError' exception if that is not possible.
    """
    while a<amin:
        a=a+2.0*numpy.pi
    while a>amax:
        a=a-2.0*numpy.pi
    if amin<=a<=amax:
        return a
    else:
        raise ValueError('angle cannot be moved into range')

def rangeastype(p1,p2,type,crit=0.001*projtls.degrees):
    """ Convert positions 'p1' and 'p2' to 'type', keeping
        differences in axes angles.
    """
    import math
    if p1.type!=p2.type:
        raise TypeError('need position objects of same type')
    q1=p1.asType(type)
    q2=p2.asType(type)
    for name in ('theta','omega','phi','kappa','chi'):
        if hasattr(p1,name) and hasattr(p2,name) and hasattr(q1,name) and hasattr(q2,name):
            start=getattr(p1,name)
            end=getattr(p2,name)
            while (getattr(q2,name)-getattr(q1,name))-(end-start)>+crit:
                setattr(q2,name,getattr(q2,name)-2.0*math.pi)
            while (getattr(q2,name)-getattr(q1,name))-(end-start)<-crit:
                setattr(q2,name,getattr(q2,name)+2.0*math.pi)
    return q1,q2

def Position(p=None,type=None,**kw):
    """ Position factory function.

        parameters:
       
            **type** -- 'k' for kappa type, 'e' for Euler, 'f' for horizonal
                        Euler, 'n' for euler with counter-clock-wise motors
                        except phi, 'g' for goniometermatrix, 'b' for bisecting,
                        'c' for C-vector, 'x' for X8 kappa, 't' for theta1-theta2.
                        'd' for horizontal euler like in MAR DTB
               
            **hardware** -- hardware used for checking possibilities and for
                            calculations using kappa support angle and wavelength.
                            This can be omitted, but this makes it very difficult
                            to perform conversions.
            
        optional (type specific) parameters::
       
            if type='k': omega, kappa, dx, phi, theta
            if type='e': omega, chi, dx, phi, theta
            if type='f': omega, chi, dx, phi, theta
            if type='d': omega, chi, dx, phi, theta
            if type='n': omega, chi, dx, phi, theta
            if type='b': dx, theta, phi, chi, psi
            if type='c': dx, c, psi, signtheta
            if type='g': dx, theta, matrix
            if type='x': omega, kappa, dx, phi, theta
            if type='t': theta1, theta2, dx
            

        matrix= 3x3-matrix, c= 3-vector, dx in mm, angles in radians.

        Alternatively, a position object 'p' can be passed, and a copy
        will be returned.
    """
    if p:
        return p.__class__(p)
    elif type in typelist:
        return typelist[type](**kw)
    else:
        raise TypeError("unknown type specified in Position()")

class PositionBase:
    """ Base class for all position types.

        All subclasses must define:

        **type** -- a 1-character type identifier for the position.

        **__init__()** -- constructor, taking all angle/vector/matrix arguments
                          plus the 'hardware'. Should store all of the
                          recognized arguments as attributes, and complain
                          about others.  It should also accept to be called
                          with only a 'p' argument, referring to a position of
                          the same type, and make itself a copy of 'p'.
             
        **With(x=y)** -- a 'modifier' method that returns the position with
                         one or more of the attributes changed. Position should
                         be 'immutable', so this method must return a copied
                         object!

        **as_?()** -- conversion routines for ? in all types (but capitalized)
                      (Currently 'as_B()', 'as_C()', 'as_E()', 'as_F()',
                      'as_G()', 'as_K()', 'as_N()', 'as_X()') including itself.
                      Many of these can be implemented as cascades.

        **asType(type=?) -- conversion routine that takes the target type as
                            a string parameter.

        **towards(other,fraction)** -- interpolation function that takes an
                      'other' position of the same type, and an interpolation
                      fraction, and that returns a position that can be
                      mathematically expressed as::

                      self*(1-fraction)+other*fraction

        **Storable()** -- return a dictionary with all arguments to the
                          constructor that can be used to reconstruct the
                          position object including the 'type', but excluding
                          the 'hardware'.
             
        **__repr__()** -- a unique string representation, presentable to
                          the user.

        Some methods are specific to specific 'type's of positions:

        **Alternate()** -- for kappa and eulerian positions, returns a position
                           with different angles but the same orientation
                           of the crystal on the goniostat.

        **NegateTheta()** -- for 'k' positions, returns a position with
                             theta negative that has the same diffraction
                             vector active at the center of the detector.

        **Xrot(angle)** -- for 'g' positions, rotates the crystal 'angle'
                            clockwise around the X axis.

        **Yrot(angle)** -- for 'g' positions, rotates the crystal 'angle'
                            clockwise around the Y axis.

        **Zrot(angle)** -- for 'g' positions, rotates the crystal 'angle'
                            clockwise around the Z axis.
    """
    def detector_edgevec(self,angle,margin=0):
        """ A vector pointing to the edge of the detector.

            Uses the DetectorEdgeMM method of the hardware, and has
            the same parameters.
        """
        mmx,mmy=self.hardware.DetectorEdgeMM(angle,margin)
        v=numpy.array([-self.dx,mmx,mmy])
        rv=matrix.matvec(matrix.Zrot(2.0*self.theta),v)
        return rv

    def detector_edgetheta(self,angle,margin=0):
        """ The theta corresponding to the detector_edgevec

            Note: this is an "absolute value"!
        """
        v=self.detector_edgevec(angle,margin)
        return matrix.vectorangle(v,self.hardware.s0)

    def detector_hitrotvec(self,v,margin=0):
        """ See whether vector 'v' at GonZero hits the
            detector in the current position.
        """
        g=self.goniometermatrix()
        v=matrix.matvec(g,v)
        return self.detector_hitvec(v,margin=margin)
    
    def detector_hitvec(self,v,margin=0):
        """ See whether vector 'v' hits the detector
            in the current position.
        """
        # Rotate back to detector swing 0
        v=matrix.matvec(matrix.Zrot(2.0*self.theta),v)
        # Scale
        f=(-self.dx/min(v[0],-0.00001))
        v=v*f
        return self.hardware.Hit(v[1],v[2],margin=margin)
    
    def detector_twothetaspan(self,margin=0):
        """ Calculate the effective coverage of the
            theta range (including a "sign").
        """
        t1l=self.detector_edgetheta(150*projtls.degrees,margin=margin)
        t2l=self.detector_edgetheta(180*projtls.degrees,margin=margin)
        t3l=self.detector_edgetheta(210*projtls.degrees,margin=margin)
        tl=min(t1l,t2l,t3l)
        vl=self.detector_edgevec(180*projtls.degrees,margin=margin)
        if vl[1]<0:
            tl=-tl
        t1r=self.detector_edgetheta(-30*projtls.degrees,margin=margin)
        t2r=self.detector_edgetheta(     0*projtls.degrees,margin=margin)
        t3r=self.detector_edgetheta( 30*projtls.degrees,margin=margin)
        tr=min(t1r,t2r,t3r)
        vr=self.detector_edgevec(  0*projtls.degrees,margin=margin)
        if vr[1]<0:
            tr=-tr
        return tl,tr

    def WithLowTwoTheta(self,ang,margin=0):
        """ Calculate a position that has the left-effective theta range ang.
        """
        p=self
        left,right=p.detector_twothetaspan(margin=margin)
        brake=0
        while abs(ang-left)>0.01*projtls.degrees and brake<5:
            brake+=1
            # Bram had a problem where a proposed scan in PhiStrategy was going for
            # 362 and 4xx degrees in theta. Try to fix that using normalangle?!?
            p=p.With(theta=normalangle(p.theta+(ang-left)/2))
            left,right=p.detector_twothetaspan(margin=margin)
        return p

    def WithHighTwoTheta(self,ang,margin):
        """ Calculate a position that has the right-effective theta range ang.
        """
        p=self
        left,right=p.detector_twothetaspan(margin=margin)
        brake=0
        while abs(ang-right)>0.01*projtls.degrees and brake<5:
            brake+=1
            # Bram had a problem where a proposed scan in PhiStrategy was going for
            # 362 and 4xx degrees in theta. Try to fix that using normalangle?!?
            p=p.With(theta=normalangle(p.theta+(ang-right)/2))
            left,right=p.detector_twothetaspan(margin=margin)
        return p

    def rotateBackTwoTheta(self,v):
        if self.theta==0:
            return v
        # For vertical theta axes.
        if not hasattr(self,'_tttmat'):
            if self.theta_clockwise:
                self._tttmat=matrix.Zrot(-2.0*self.theta)
            else:
                self._tttmat=matrix.Zrot(2.0*self.theta)
        return matrix.matvec(self._tttmat,v)

    def rotateToTwoTheta(self,v):
        if self.theta==0:
            return v
        # For vertical theta axes.
        if not hasattr(self,'_bttmat'):
            if self.theta_clockwise:
                self._bttmat=matrix.Zrot(2.0*self.theta)
            else:
                self._bttmat=matrix.Zrot(-2.0*self.theta)
        return matrix.matvec(self._bttmat,v)
        
    def goniometermatrix(self):
        """ The goniometer orientation matrix equivalent to this position.
        """
        return self.as_G().matrix

    def oranges(self):
        """ The allowed omega ranges for the position on a Kappa Goniostat.

            Disregards the current value of omega.

            Returns a n-tuple of 2-tuples.
        """
        if self.hardware:
            return self.hardware.GetOmegaRanges(self)
        else:
            return ((-numpy.pi,numpy.pi),)

    def orange(self):
        """ The allowed omega ranges for the position on a Kappa Goniostat.

            Takes the current value of omega into account.

            Returns a n-tuple of tuples.
        """
        if self.hardware:
            return self.hardware.GetOmegaRange(self)
        else:
            return -numpy.pi,numpy.pi

    def allowed(self):
        """ Check whether the position is allowed on the goniostat.
        """
        if self.hardware:
            return self.hardware.PosHardAllowed(self)
        else:
            return 1

    def asType(self,type):
        """ Return a new 'Position()' object of type 'type' that
            represents this position.
        """
        type=type[:1].upper()
        return getattr(self,'as_%c'%type)()

class _KPosition(PositionBase):
    type='k'
    theta_clockwise=1
    phi_clockwise=1
    omega_clockwise=1
    def __init__(self,hardware=None,p=None,dx=None,theta=None,omega=None,kappa=None,phi=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.omega=p.omega
            self.kappa=p.kappa
            self.phi=normalangle(p.phi)
        else:
            self.dx=dx
            self.theta=theta
            self.omega=omega
            self.kappa=kappa
            self.phi=phi

    def toDatabase(self,dbobj):
        """ Write position to database.
        """
        import projtls
        dbobj['position_type']='k'
        dbobj['dx']=self.dx
        dbobj['theta']=self.theta/projtls.degrees*2
        dbobj['omega']=self.omega/projtls.degrees
        dbobj['kappa']=self.kappa/projtls.degrees
        dbobj['phi']=self.phi/projtls.degrees

    def Storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'k',
            'dx':self.dx,
            'theta':self.theta,
            'omega':self.omega,
            'kappa':self.kappa,
            'phi':self.phi}

    def Alternate(self):
        """ The alternate Kappa position that has the same orientation.
        """
        return self.as_E().Alternate().as_K()

    def NegateTheta(self):
        """ The same position with negative theta.

            Keep the same reflection active in the center of the detector.
        """
        return self.With(theta=-self.theta,omega=self.omega+180*projtls.degrees-2*self.theta)

    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_F()

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_D()
    
    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if not self.hardware:
            raise error("Cannot perform conversion without knowing goniometer")
        if not hasattr(self,'alpha'):
            self.alpha=self.hardware.alpha
        return Position(dx=self.dx,
            hardware=self.hardware,
            type='g',
            theta=self.theta,
            matrix=numpy.dot(matrix.Zrot(self.omega),
                numpy.dot(matrix.Yrot(-self.alpha),
                numpy.dot(matrix.Zrot(self.kappa),
                numpy.dot(matrix.Yrot(self.alpha),matrix.Zrot(self.phi))))))

    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if not self.hardware:
            raise error("Cannot perform conversion without knowing goniometer")
        if not hasattr(self,'alpha'):
            self.alpha=self.hardware.alpha
        if self.kappa is None:
            print("DBG> Convert incomplete kappa to eulerian!")
            return Position(type='e',
                hardware=self.hardware,
                theta=self.theta,
                dx=self.dx)
        halfkappa = 0.5 * self.kappa
        # Modulo 360
        while halfkappa > numpy.pi/2:
            halfkappa -= numpy.pi
        while halfkappa < -numpy.pi/2:
            halfkappa += numpy.pi
        sinx=numpy.cos(self.alpha)*numpy.sin(halfkappa)
        cosx=numpy.cos(halfkappa)
        x=numpy.arctan2(sinx,cosx)
        omegae=self.omega+x
        phie=self.phi+x
        sinc=numpy.sin(self.alpha)*numpy.sin(halfkappa)
        chie=2.0*numpy.arcsin(sinc)
        return Position(type='e',
            hardware=self.hardware,
            theta=self.theta,
            omega=normalangle(omegae),
            chi=normalangle(chie),
            phi=normalangle(phie),
            dx=self.dx)

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_B()

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_B().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_N()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if self.omega is not None:
            om=180*projtls.degrees-self.omega
        else:
            om=None
        if self.phi is not None:
            ph=self.phi-90*projtls.degrees
        else:
            ph=None
        if self.theta is not None:
            th=-self.theta
        else:
            th=None
        return Position(type='x',
            hardware=self.hardware,
            dx=self.dx,theta=th,omega=om,kappa=self.kappa,phi=ph)
    
    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_T()
        
   
    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='k',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            omega=kw.get('omega',self.omega),
            kappa=kw.get('kappa',self.kappa),
            phi=kw.get('phi',self.phi))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        kw={}
        if self.dx is not None and other.dx is not None:
            kw['dx']=self.dx*f0+other.dx*f1
        if self.theta is not None and other.theta is not None:
            kw['theta']=self.theta*f0+other.theta*f1
        if self.omega is not None and other.omega is not None:
            kw['omega']=self.omega*f0+other.omega*f1
        if self.phi is not None and other.phi is not None:
            kw['phi']=self.phi*f0+other.phi*f1
        if self.kappa is not None and other.kappa is not None:
            kw['kappa']=self.kappa*f0+other.kappa*f1
        if self.hardware is None and other.hardware is not None:
            kw['hardware']=other.hardware
        return self.With(**kw)
    
    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[Kappa angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.omega is not None:
            s=s+" omega=%8.3f"%(self.omega/projtls.degrees)
        if self.kappa is not None:
            s=s+" kappa=%8.3f"%(self.kappa/projtls.degrees)
        s=s+"]"
        return s

class _XPosition(PositionBase):
    type='x'
    theta_clockwise=0
    phi_clockwise=1
    omega_clockwise=0
    def __init__(self,hardware=None,p=None,dx=None,theta=None,omega=None,kappa=None,phi=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.omega=p.omega
            self.kappa=p.kappa
            self.phi=normalangle(p.phi)
        else:
            self.dx=dx
            self.theta=theta
            self.omega=omega
            self.kappa=kappa
            self.phi=phi

    def toDatabase(self,dbobj):
        """ Write position to database.
        """
        import projtls
        dbobj['position_type']='x'
        dbobj['dx']=self.dx
        dbobj['theta']=self.theta/projtls.degrees*2
        dbobj['omega']=self.omega/projtls.degrees
        dbobj['kappa']=self.kappa/projtls.degrees
        dbobj['phi']=self.phi/projtls.degrees

    def storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'x',
            'dx':self.dx,
            'theta':self.theta,
            'omega':self.omega,
            'kappa':self.kappa,
            'phi':self.phi}

    def as_D(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_D()

    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_F()
    
    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_G()

    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_K().as_E()

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_B()

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if self.theta is not None:
            th=-self.theta
        else:
            th=None
        if self.omega is not None:
            om=180*projtls.degrees-self.omega
        else:
            om=None
        if self.phi is not None:
            ph=self.phi+90*projtls.degrees
        else:
            ph=None
        return Position(type='k',
                        hardware=self.hardware,
                        dx=self.dx,theta=th,omega=om,kappa=self.kappa,phi=ph)

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_N()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_T(self):
        return self.as_N().as_T()

    def With(self,**kw):
        """ Make clone of this position With some angle(s) changed.
        """
        return Position(type='x',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            omega=kw.get('omega',self.omega),
            kappa=kw.get('kappa',self.kappa),
            phi=kw.get('phi',self.phi))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        kw={}
        if self.dx is not None and other.dx is not None:
            kw['dx']=self.dx*f0+other.dx*f1
        if self.theta is not None and other.theta is not None:
            kw['theta']=self.theta*f0+other.theta*f1
        if self.omega is not None and other.omega is not None:
            kw['omega']=self.omega*f0+other.omega*f1
        if self.phi is not None and other.phi is not None:
            kw['phi']=self.phi*f0+other.phi*f1
        if self.kappa is not None and other.kappa is not None:
            kw['kappa']=self.kappa*f0+other.kappa*f1
        if self.hardware is None and other.hardware is not None:
            kw['hardware']=other.hardware
        return self.With(**kw)

    def Alternate(self):
        """ The Alternate Kappa position that has the same orientation.
        """
        return self.as_E().Alternate().as_X()

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[X8 Kappa angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.omega is not None:
            s=s+" omega=%8.3f"%(self.omega/projtls.degrees)
        if self.kappa is not None:
            s=s+" kappa=%8.3f"%(self.kappa/projtls.degrees)
        s=s+"]"
        return s

class _BPosition(PositionBase):
    type='b'
    theta_clockwise=1
    phi_clockwise=1
    omega_clockwise=1
    def __init__(self,hardware=None,p=None,dx=None,theta=None,phi=None,chi=None,psi=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if hardware:
            self.wavelength=hardware.wavelength
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.phi=p.phi
            self.chi=p.chi
            self.psi=p.psi
        else:
            self.dx=dx
            self.theta=theta
            self.phi=phi
            self.chi=chi
            self.psi=psi

    def toDatabase(self,dbobj):
        """ Write position to database.
        """

    def storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'b',
            'dx':self.dx,
            'theta':self.theta,
            'phi':self.phi,
            'chi':self.chi,
            'psi':self.psi}

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if not self.hardware:
            raise error("Cannot perform conversion without knowing wavelength")
        if self.theta>=0:
            signtheta=1
        else:
            signtheta=-1
        d=2*numpy.sin(self.theta)/self.wavelength
        c=[-numpy.sin(self.phi)*numpy.cos(self.chi)*d,
            numpy.cos(self.phi)*numpy.cos(self.chi)*d,
            numpy.sin(self.chi)*d]
        return Position(type='c',hardware=self.hardware,dx=self.dx,c=c,psi=self.psi,signtheta=signtheta)

    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        sinpsi=numpy.sin(self.psi)
        cospsi=numpy.cos(self.psi)
        signth=sign(numpy.sin(self.theta))
        sinchb=numpy.sin(self.chi)
        coschb=numpy.cos(self.chi)
        signcb=sign(coschb)
        sinche=signcb*sign(sinchb)*numpy.sqrt((cospsi*sinchb)**2+sinpsi**2)
        cosche=signth*signcb*cospsi*coschb
        try:
            chi=numpy.arctan2(sinche,cosche)
        except ValueError:
            print("B-E Chi problem:",self)
            chi=0.0
        if sinchb==0 and sinpsi==0:
            omega=self.theta-90.0*(signcb-1.0)
            sinphe=-numpy.sin(omega-self.theta)*coschb
            cosphe=signth*signcb*numpy.cos(omega-self.theta)*cospsi
        else:
            signch=sign(numpy.sin(chi))
            sinome=signch*signcb*coschb*sinpsi
            cosome=signch*sinchb
            try:
                omega=numpy.arctan2(sinome,cosome)+self.theta
            except ValueError:
                print("Oops:",self)
                omega=0.0
            sinphe=-signch*signcb*sinpsi
            cosphe=signch*signth*signcb*sinchb*cospsi
        try:
            phi=numpy.arctan2(sinphe,cosphe)+self.phi
        except ValueError:
            print("Oops:",self)
            phi=0.0
        return Position(type='e',hardware=self.hardware,dx=self.dx,
            theta=self.theta,
            omega=normalangle(omega),
            chi=normalangle(chi),
            phi=normalangle(phi))

    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_F()

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_D()

    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if self.theta>0:
            S=0
        else:
            S=numpy.pi
        if numpy.cos(self.chi)>0:
            C=0
        else:
            C=numpy.pi
        return Position(dx=self.dx,
            hardware=self.hardware,
            type='g',
            theta=self.theta,
            matrix=numpy.dot(matrix.Zrot(self.theta),
                numpy.dot(matrix.Xrot(S),
                numpy.dot(matrix.Yrot(-self.psi),
                numpy.dot(matrix.Zrot(S+C),
                numpy.dot(matrix.Xrot(C),
                numpy.dot(matrix.Xrot(self.chi),matrix.Zrot(self.phi))))))))

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_K()

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_N()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_K().as_X()
        
    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_T()
        

    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='b',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            phi=kw.get('phi',self.phi),
            chi=kw.get('chi',self.chi),
            psi=kw.get('psi',self.psi))

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[Bisecting angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.chi is not None:
            s=s+" chi=%8.3f"%(self.chi/projtls.degrees)
        if self.psi is not None:
            s=s+" psi=%8.3f"%(self.psi/projtls.degrees)
        s=s+"]"
        return s

class _CPosition(PositionBase):
    type='c'
    theta_clockwise=1
    def __init__(self,hardware=None,p=None,dx=None,c=None,psi=None,signtheta=1):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if hardware:
            self.wavelength=hardware.wavelength
        if p:
            self.dx=p.dx
            self.c=(p.c[0],p.c[1],p.c[2])
            self.psi=p.psi
            self.signtheta=p.signtheta
        else:
            self.dx=dx
            self.c=(c[0],c[1],c[2])
            if psi is None:
                self.psi=0.0
            else:
                self.psi=psi
            self.signtheta=signtheta

    def toDatabase(self,dbobj):
        """ Write position to database.
        """

    def storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'k',
            'dx':self.dx,
            'c':self.c,
                'signtheta':self.signtheta,
            'psi':self.psi}

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if not self.hardware:
            raise error("Cannot perform conversion without knowing wavelength")
        cosx=numpy.sqrt(self.c[0]**2+self.c[1]**2)
        chi=numpy.arctan2(self.c[2],cosx)
        if cosx<1.e-6:
            phi=0.0
        else:
            try:
                phi=numpy.arctan2(-self.c[0],self.c[1])
            except ValueError:
                print("Oops: ",self)
                phi=0
        sinx=numpy.sqrt(cosx**2+self.c[2]**2)*self.wavelength/2.0
        if sinx>=1.0:
            theta=self.signtheta*numpy.pi/2.0
        else:
            theta=self.signtheta*numpy.arcsin(sinx)
        if self.signtheta<0:
            #print("DBG> C2B Theta negative",theta/projtls.degrees)
            phi=phi+180*projtls.degrees
            chi=-chi
        return Position(type='b',
            hardware=self.hardware,
            dx=self.dx,
            theta=normalangle(theta),
            phi=normalangle(phi),
            chi=normalangle(chi),
            psi=self.psi)

    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_B().as_E()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_K()

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_F()

    def as_D(self):
        """Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_D()

    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_G()

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_N()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_K().as_X()

    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_T()
        

    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='c',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            c=kw.get('c',self.c),
                        signtheta=kw.get('signtheta',self.signtheta),
            psi=kw.get('psi',self.psi))

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        if self.psi is not None:
            psi="%8.3f"%(self.psi/projtls.degrees)
        else:
            psi="None"
        if self.dx is not None:
            dx="%6.2f"%self.dx
        else:
            dx="None"
        return "[C-vector: dx=%s c=%s psi=%s]"%(dx,repr(self.c),psi)

class _EPosition(PositionBase):
    type='e'
    theta_clockwise=1
    phi_clockwise=1
    omega_clockwise=1
    def __init__(self,hardware=None,p=None,dx=None,theta=None,omega=None,chi=None,phi=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.omega=p.omega
            self.chi=p.chi
            self.phi=p.phi
        else:
            self.dx=dx
            self.theta=theta
            self.omega=omega
            self.chi=chi
            self.phi=phi

    def toDatabase(self,dbobj):
        """ Write position to database.
        """
        import projtls
        dbobj['position_type']='e'
        dbobj['dx']=self.dx
        dbobj['theta']=self.theta/projtls.degrees*2
        dbobj['omega']=self.omega/projtls.degrees
        dbobj['chi']=self.chi/projtls.degrees
        dbobj['phi']=self.phi/projtls.degrees

    def storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'e',
            'dx':self.dx,
            'theta':self.theta,
            'omega':self.omega,
            'chi':self.chi,
            'phi':self.phi}

    def Alternate(self):
        """ The alternate Eulerian position that has the same orientation.
        """
        return Position(type='e',
            hardware=self.hardware,
            dx=self.dx,
            omega=normalangle(self.omega+numpy.pi),
            theta=self.theta,
            chi=-self.chi,
            phi=normalangle(self.phi+numpy.pi))

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        sineps=numpy.sin(self.omega-self.theta)
        coseps=numpy.cos(self.omega-self.theta)
        if coseps<0:
            signcb=-1.0
        else:
            signcb=1.0
        coschi=numpy.cos(self.chi)
        sinchi=numpy.sin(self.chi)
        if self.theta<0:
            signth=-1.0
        else:
            signth=1.0
        sinchib=coseps*sinchi
        coschib=signcb*numpy.sqrt(coschi**2+(sinchi*sineps)**2)
        chib=numpy.arctan2(sinchib,coschib)
        if sineps==0 and coschi==0:
            phib=0
            sinpsi=-signcb*numpy.sin(self.phi)*sinchi
            cospsi=signth*signcb*numpy.cos(self.phi)*coseps
        else:
            sinphi=-signcb*sineps
            cosphi=signcb*coseps*coschi
            phib=self.phi-numpy.arctan2(sinphi,cosphi)
            sinpsi=sinchi*sineps
            cospsi=signth*coschi
        psi=numpy.arctan2(sinpsi,cospsi)
        return Position(type='b',
            hardware=self.hardware,
            dx=self.dx,
            theta=self.theta,
            phi=normalangle(phib),
            chi=normalangle(chib),
            psi=normalangle(psi))

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_B().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if not self.hardware:
            raise error("Cannot perform conversion without knowing goniometer")
        if not hasattr(self,'alpha'):
            self.alpha=self.hardware.alpha
            self.kappamax=min(self.hardware.kappamax,-self.hardware.kappamin)
        si=numpy.sin(0.5*self.chi)
        co=numpy.sin(self.alpha)**2-si**2
        con3=numpy.cos(self.kappamax/2.0)**2 * numpy.sin(self.alpha)**2
        if con3>co:
            kappa=180.0*projtls.degrees
            omega=90.0*projtls.degrees
            raise error("Chi high")
        else:
            co=numpy.sqrt(co)
            kappa=2.0*numpy.arctan2(si,co)
            si=si*numpy.cos(self.alpha)
            omega=numpy.arctan2(si,co)
        phi=self.phi-omega
        omega=-omega+self.omega
        return Position(type='k',
            hardware=self.hardware,
            dx=self.dx,
            theta=self.theta,
            omega=normalangle(omega),
            kappa=normalangle(kappa),
            phi=normalangle(phi))
    
    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_F()

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_D()

    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        #print(>> sys.stderr, self)
        return Position(dx=self.dx,
            hardware=self.hardware,
            type='g',
            theta=self.theta,
            matrix=numpy.dot(matrix.Zrot(self.omega),
                numpy.dot(matrix.Xrot(self.chi),matrix.Zrot(self.phi))))

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if self.phi is not None:
            ph=self.phi-90*projtls.degrees
        else:
            ph=None
        if self.omega is not None:
            om=180*projtls.degrees-self.omega
        else:
            om=None
        if self.theta is not None:
            th=-self.theta
        else:
            th=None
        return Position(hardware=self.hardware,
                        type='n',
                        dx=self.dx,theta=th,chi=self.chi,phi=ph,omega=om)

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_K().as_X()
    
    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_T()

    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='e',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            omega=kw.get('omega',self.omega),
            chi=kw.get('chi',self.chi),
            phi=kw.get('phi',self.phi))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        return self.With(dx=self.dx*f0+other.dx*f1,
            theta=self.theta*f0+other.theta*f1,
            omega=self.omega*f0+other.omega*f1,
            chi=self.chi*f0+other.chi*f1,
            phi=self.phi*f0+other.phi*f1)

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[Eulerian angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.omega is not None:
            s=s+" omega=%8.3f"%(self.omega/projtls.degrees)
        if self.chi is not None:
            s=s+" chi=%8.3f"%(self.chi/projtls.degrees)
        s=s+"]"
        return s

class _FPosition(PositionBase):
    """ Horizontal euler goniostat.
    """
    type='f'
    theta_clockwise=0
    phi_clockwise=0
    omega_clockwise=0
    def __init__(self,hardware=None,p=None,dx=None,theta=None,omega=None,chi=None,phi=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.omega=p.omega
            self.chi=p.chi
            self.phi=p.phi
        else:
            self.dx=dx
            self.theta=theta
            self.omega=omega
            self.chi=chi
            self.phi=phi

    def toDatabase(self,dbobj):
        """ Write position to database.
        """

    def Storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'f',
            'dx':self.dx,
            'theta':self.theta,
            'omega':self.omega,
            'chi':self.chi,
            'phi':self.phi}

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_B()

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_K()
    
    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_G().as_E()
    
    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_D()

    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return Position(dx=self.dx,
            hardware=self.hardware,
            type='g',
            theta=self.theta,
            matrix=numpy.dot(matrix.Xrot(90*projtls.degrees),
                numpy.dot(matrix.Zrot(self.omega),
                numpy.dot(matrix.Xrot(self.chi),matrix.Zrot(self.phi)))))

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_N()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_X()

    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_T()
        

    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='f',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            omega=kw.get('omega',self.omega),
            chi=kw.get('chi',self.chi),
            phi=kw.get('phi',self.phi))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        return self.With(dx=self.dx*f0+other.dx*f1,
            theta=self.theta*f0+other.theta*f1,
            omega=self.omega*f0+other.omega*f1,
            chi=self.chi*f0+other.chi*f1,
            phi=self.phi*f0+other.phi*f1)

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[Horizontal Eulerian angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.omega is not None:
            s=s+" omega=%8.3f"%(self.omega/projtls.degrees)
        if self.chi is not None:
            s=s+" chi=%8.3f"%(self.chi/projtls.degrees)
        s=s+"]"
        return s

class _DPosition(PositionBase):
    """Horizontal euler goniostat like in the MAR DTB
    """
    type='d'
    theta_clockwise=0
    omega_clockwise=0
    phi_clockwise=0
    phizero=0#-90*projtls.degrees
    def __init__(self,hardware=None,p=None,dx=None,theta=None,omega=None,chi=None,phi=None):
        """Constructor. Part of Position subclass protocol
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.omega=p.omega
            self.chi=p.chi
            self.phi=p.phi
        else:
            self.dx=dx
            self.theta=theta
            self.omega=omega
            self.chi=chi
            self.phi=phi

    def Storable(self):
        """Dictionary representation. Part of Position subclass protocol
        """
        return {'type':'d',
                'dx':self.dx,
                'theta':self.theta,
                'omega':self.omega,
                'chi':self.chi,
                'phi':self.phi}

    def as_B(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_B()

    def as_C(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_C()

    def as_K(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_K()

    def as_X(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_K().as_X()
    
    def as_E(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_E()
    
    def as_F(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_F()

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.With()

    def as_G(self):
        """Conversion. Part of Position subclass protocol
        """
        return Position(dx=self.dx,
                        hardware=self.hardware,
                        type='g',
                        theta=self.theta,
                        matrix=numpy.dot(matrix.Xrot(-90*projtls.degrees),
                               numpy.dot(matrix.Zrot(self.omega),
                               numpy.dot(matrix.Xrot(self.chi),matrix.Zrot(self.phi-self.phizero)))))

    def as_N(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_E().as_N()

    def With(self,**kw):
        """Make clone of this position with some angle(s) changed."""
        return Position(type='d',
                        hardware=kw.get('hardware',self.hardware),
                        dx=kw.get('dx',self.dx),
                        theta=kw.get('theta',self.theta),
                        omega=kw.get('omega',self.omega),
                        chi=kw.get('chi',self.chi),
                        phi=kw.get('phi',self.phi))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        return self.With(dx=self.dx*f0+other.dx*f1,
                        theta=self.theta*f0+other.theta*f1,
                        omega=self.omega*f0+other.omega*f1,
                        chi=self.chi*f0+other.chi*f1,
                        phi=self.phi*f0+other.phi*f1)

    def __repr__(self):
        """Representation. Part of Position subclass protocol
        """
        s="[Eulerian DTB angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.omega is not None:
            s=s+" omega=%8.3f"%(self.omega/projtls.degrees)
        if self.chi is not None:
            s=s+" chi=%8.3f"%(self.chi/projtls.degrees)
        s=s+"]"
        return s

class _NPosition(PositionBase):
    """ Counter-clockwise rotating eulerian goniostat, phi clockwise!
    """
    type='n'
    theta_clockwise=0
    phi_clockwise=1
    omega_clockwise=0
    def __init__(self,hardware=None,p=None,dx=None,theta=None,omega=None,chi=None,phi=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.omega=p.omega
            self.chi=p.chi
            self.phi=p.phi
        else:
            self.dx=dx
            self.theta=theta
            self.omega=omega
            self.chi=chi
            self.phi=phi

    def toDatabase(self,dbobj):
        """ Write position to database.
        """
        import projtls
        dbobj['position_type']='n'
        dbobj['dx']=self.dx
        dbobj['theta']=self.theta/projtls.degrees*2
        dbobj['omega']=self.omega/projtls.degrees
        dbobj['chi']=self.chi/projtls.degrees
        dbobj['phi']=self.phi/projtls.degrees

    def Storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'n',
            'dx':self.dx,
            'theta':self.theta,
            'omega':self.omega,
            'chi':self.chi,
            'phi':self.phi}

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_B()

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_K()
    
    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        if self.omega is not None:
            om=180*projtls.degrees-self.omega
        else:
            om=None
        if self.phi is not None:
            ph=self.phi+90*projtls.degrees
        else:
            ph=None
        if self.theta is not None:
            th=-self.theta
        else:
            th=None
        return Position(hardware=self.hardware,
                        type='e',
                        dx=self.dx,theta=th,chi=self.chi,phi=ph,omega=om)
    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_F()

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        return self.as_G().as_D()
    
    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_G()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_X()

    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        th1 = self.omega
        th2 = self.theta*2 - self.omega
        
        return Position(type='t',
            hardware = self.hardware,
            dx=self.dx,
            theta1=th1,
            theta2=th2)
        
        
        
    
    def Alternate(self):
        """ The alternate N-position that has the same orientation.
        """
        return Position(type='n',
            hardware=self.hardware,
            dx=self.dx,
            omega=normalangle(self.omega+numpy.pi),
            theta=self.theta,
            chi=-self.chi,
            phi=normalangle(self.phi+numpy.pi))

    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='n',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            omega=kw.get('omega',self.omega),
            chi=kw.get('chi',self.chi),
            phi=kw.get('phi',self.phi))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        return self.With(dx=self.dx*f0+other.dx*f1,
            theta=self.theta*f0+other.theta*f1,
            omega=self.omega*f0+other.omega*f1,
            chi=self.chi*f0+other.chi*f1,
            phi=self.phi*f0+other.phi*f1)

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[Counterclockwise Eulerian angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta is not None:
            s=s+" theta=%8.3f"%(self.theta/projtls.degrees)
        if self.phi is not None:
            s=s+" phi=%8.3f"%(self.phi/projtls.degrees)
        if self.omega is not None:
            s=s+" omega=%8.3f"%(self.omega/projtls.degrees)
        if self.chi is not None:
            s=s+" chi=%8.3f"%(self.chi/projtls.degrees)
        s=s+"]"
        return s

class _GPosition(PositionBase):
    type='g'
    theta_clockwise=1
    def __init__(self,hardware=None,p=None,dx=None,matrix=None,theta=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta=p.theta
            self.matrix=p.matrix
        else:
            self.dx=dx
            self.theta=theta
            self.matrix=matrix

    def toDatabase(self,dbobj):
        """ Write position to database.
        """

    def Storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'g',
            'dx':self.dx,
            'theta':self.theta,
            'matrix':self.matrix}

    def as_E(self,zeromega=0):
        """ Conversion. Part of Position subclass protocol.
        """
        # The goniometermatrix from the eulerian angles= 
        # Zrot(omega)*Xrot(chi)*Zrot(phi)
        # So, with a=cos(phi)
        #     b=sin(phi)
        #     c=cos(chi)
        #     d=sin(chi)
        #     e=cos(omega)
        #     f=sin(omega)
        # The goniometermatrix matrix[0..2][0..2] is:
        #   ( a -b  0 )   ( 1  0  0 )   ( e -f  0 )
        #   ( b  a  0 ) * ( 0  c -d ) * ( f  e  0 )
        #   ( 0  0  1 )   ( 0  d  c )   ( 0  0  1 )
        # or:
        #   ( a -b  0 )   ( e   -f  0 )
        #   ( b  a  0 ) * ( fc  ec -d )
        #   ( 0  0  1 )   ( fd  ed  c )
        # or:
        #   ( ae-bfc  -fa-bec   bd )
        #   ( be+afc  -fb+aec  -ad )
        #   (  fd        ed      c )
        # or:
        coschi=self.matrix[2,2]
        # Assume chi positive.
        try:
            sinchi=numpy.sqrt(1-coschi**2)
        except ValueError:
            sinchi=0.0
        chi=numpy.arctan2(sinchi,coschi)
        if abs(sinchi)<1.e-10:
            # Choose omega to be 180 degrees (easily accessible)
            sinomega=0.0
            if zeromega:
                cosomega=1.0
            else:
                cosomega=-1.0
        else:
            sinomega=self.matrix[0,2]/sinchi
            cosomega=self.matrix[1,2]/sinchi
        omega=numpy.arctan2(sinomega,cosomega)
        if abs(sinchi)<1.e-10:
            # Since f==0 and e==-1, the solution is simple....
            sinphi=self.matrix[1,0]
            cosphi=-self.matrix[0,0]
        else:
            sinphi=self.matrix[2,0]/sinchi
            cosphi=-self.matrix[2,1]/sinchi
        phi=numpy.arctan2(sinphi,cosphi)
        return Position(type='e',
            hardware=self.hardware,
            dx=self.dx,
            theta=self.theta,
            omega=omega,
            phi=phi,
            chi=chi)

    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        t=self.Xrot(-90*projtls.degrees).as_E(zeromega=1)
        return Position(type='f',
            hardware=t.hardware,
            dx=t.dx,
            theta=t.theta,
            omega=t.omega,
            phi=t.phi,
            chi=t.chi)

    def as_D(self):
        """Conversion. Part of Position subclass protocol
        """
        t=self.Xrot(90*projtls.degrees).as_E()
        return Position(type='d',
                        hardware=t.hardware,
                        dx=t.dx,
                        theta=t.theta,
                        omega=t.omega,
                        phi=t.phi+_DPosition.phizero,
                        chi=t.chi)

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_K()
    
    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_B()
    
    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.With()

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_E().as_N()
        
    def as_T(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_T()

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_K().as_X()

    def Xrot(self,angle):
        """ Rotate 'angle' (clockwise) around the X axis.
        """
        return self.With(matrix=numpy.dot(matrix.Xrot(angle),self.matrix))

    def Yrot(self,angle):
        """ Rotate 'angle' (clockwise) around the Y axis.
        """
        return self.With(matrix=numpy.dot(matrix.Yrot(angle),self.matrix))

    def Zrot(self,angle):
        """ Rotate 'angle' (clockwise) around the Z axis.
        """
        return self.With(matrix=numpy.dot(matrix.Zrot(angle),self.matrix))

    def With(self,**kw):
        """ Make clone of this position with some angle(s) changed.
        """
        return Position(type='g',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta=kw.get('theta',self.theta),
            matrix=kw.get('matrix',self.matrix))

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        if self.theta is not None:
            theta="%8.3f"%(self.theta/projtls.degrees)
        else:
            theta="None"
        if self.dx is not None:
            dx="%6.2f"%self.dx
        else:
            dx="None"
        return "[Goniometermatrix: dx=%s theta=%s matrix=%s]"%(dx,theta,repr(self.matrix))



class _TPosition(PositionBase):
    type = 't'
    def __init__(self,hardware=None,p=None,dx=None,theta1=None,theta2=None):
        """ Constructor. Part of Position subclass protocol.
        """
        self.hardware=hardware
        if p:
            self.dx=p.dx
            self.theta1=p.theta1
            self.theta2=p.theta2
            
        else:
            self.dx=dx
            self.theta1=theta1
            self.theta2=theta2
            
    def toDatabase(self,dbobj):
        """ Write position to database.
        """
        import projtls
        dbobj['position_type']='t'
        dbobj['dx']=self.dx
        dbobj['theta1']=self.theta1/projtls.degrees
        dbobj['theta2']=self.theta2/projtls.degrees
        
    def storable(self):
        """ Dictionary representation. Part of Position subclass protocol.
        """
        return {'type':'t',
            'dx':self.dx,
            'theta1':self.theta1,
            'theta2':self.theta2}
    
    
    def With(self,**kw):
        """ Make clone of this position With some angle(s) changed.
        """
        return Position(type='t',
            hardware=kw.get('hardware',self.hardware),
            dx=kw.get('dx',self.dx),
            theta1=kw.get('theta1',self.theta1),
            theta2=kw.get('theta2',self.theta2))

    def towards(self,other,fraction):
        if not other.type==self.type:
            raise error('cannot interpolate between different typed positions')
        f0=1.0-fraction
        f1=fraction
        return self.With(dx=self.dx*f0+other.dx*f1,
            theta1=self.theta1*f0+other.theta1*f1,
            theta2=self.theta2*f0+other.theta2*f1)
            

    def __repr__(self):
        """ Representation. Part of Position subclass protocol.
        """
        s="[Counterclockwise Eulerian angles:"
        if self.dx is not None:
            s=s+" dx=%6.2f"%self.dx
        if self.theta1 is not None:
            s=s+" theta1=%8.3f"%(self.theta1/projtls.degrees)
        if self.theta2 is not None:
            s=s+" theta2=%8.3f"%(self.theta2/projtls.degrees)
        
        s=s+"]"
        
        return s
            
            
    def as_F(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_F()
    
    def as_G(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_G()

    def as_E(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_E()

    def as_B(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_B()

    def as_C(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_C()

    def as_K(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_K()

    def as_N(self):
        """ Conversion. Part of Position subclass protocol.
        """
        th = (self.theta1+self.theta2)/2
        om = self.theta1
        
        return Position(hardware=self.hardware,
                        type='n',
                        dx=self.dx,theta=th,chi=None,phi=None,omega=om)
    

    def as_X(self):
        """ Conversion. Part of Position subclass protocol.
        """
        return self.as_N().as_X()

    def as_T(self):
        return self.With()
    
################################################################################

def fromDatabase(hardware,dbobj):
    """ Read position from database.
    """
    import projtls
    kw={
        'type':dbobj['position_type'],
        'dx':dbobj['dx'],
        'theta':dbobj['theta']*projtls.degrees/2,
        'omega':dbobj['omega']*projtls.degrees,
        'phi':dbobj['phi']*projtls.degrees
    }
    if kw['type'] in 'kx':
        kw['kappa']=dbobj['kappa']*projtls.degrees
    elif kw['type'] in 'en':
        kw['chi']=dbobj['chi']*projtls.degrees

    return Position(hardware=hardware,**kw)

def PositionFromStorage(hardware,state):
    """ Return the position expressed by the dictionary 'state'.

        'state' is a dictionary created using 'position.storable()' before."""
    if state['type']=='k':
        return Position(type='k',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            omega=state['omega'],
            kappa=state['kappa'],
            phi=state['phi'])
    elif state['type']=='e':
        return Position(type='e',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            omega=state['omega'],
            chi=state['chi'],
            phi=state['phi'])
    elif state['type']=='f':
        return Position(type='f',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            omega=state['omega'],
            chi=state['chi'],
            phi=state['phi'])
    elif state['type']=='n':
        return Position(type='n',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            omega=state['omega'],
            chi=state['chi'],
            phi=state['phi'])
    elif state['type']=='g':
        return Position(type='g',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            matrix=state['matrix'])
    elif state['type']=='b':
        return Position(type='b',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            psi=state['psi'],
            chi=state['chi'],
            phi=state['phi'])
    elif state['type']=='c':
        if not 'signtheta' in state:
            state['signtheta']=1
        return Position(type='c',
            hardware=hardware,
            dx=state['dx'],
            c=state['c'],
            signtheta=state['signtheta'],
            psi=state['psi'])
    elif state['type']=='x':
        return Position(type='x',
            hardware=hardware,
            dx=state['dx'],
            theta=state['theta'],
            omega=state['omega'],
            kappa=state['kappa'],
            phi=state['phi'])
    elif state['type']=='t':
        return Position(type='t',
            hardware=hardware,
            dx=state['dx'],
            theta1=state['theta1'],
            theta2=state['theta2'])
            
    else:
        raise error("Illegal type from storage. Saved in newer version?")

def sign(a):
    """ Return -1.0 for negative numbers, 1.0 otherwise.
    """
    if a<0.0:
        return -1.0
    else:
        return 1.0

def Null(hardware):
    """ Return zero position.
    """
    return hardware.nativePosition(
        Position(type='e',hardware=hardware,theta=0,omega=0,chi=0,phi=0)
    )

def Zero(hardware=None):
    """ Return a zero-position in Kappa angles.
    """
    return Position(type='k',hardware=hardware,kappa=0,omega=0,theta=0,phi=0)

def GZero(hardware=None):
    """ Return a zero position in 'g' type.
    """
    return Position(type='g',hardware=hardware,matrix=numpy.array([[1.,0,0],[0,1,0],[0,0,1]]))

typelist={'k':_KPosition,
          'e':_EPosition,
          'f':_FPosition,
          'b':_BPosition,
          'c':_CPosition,
          'd':_DPosition,
          'g':_GPosition,
          'n':_NPosition,
          'x':_XPosition,
     #    't':_TPosition,
          }

################################################################################

if __name__=="__main__":
    pass
    from hwdriver import dummydrv
    hardware=dummydrv.DummyGon()
    # Manipulate it a bit so that we know what we're dealing With
    hardware.kappamax=178.00*projtls.degrees
    hardware.kappamin=-178.00*projtls.degrees
    hardware.wavelength=0.71073
    hardware.alpha=50.00*projtls.degrees
    # Set up a range of interesting positions
    pos=[]
    p0=Position(type='n',
            dx=50,
            hardware=hardware,
            theta=-7.23*projtls.degrees,
            omega=-10.00*projtls.degrees,
            chi=54.74*projtls.degrees,
            phi=100.00*projtls.degrees)
    pos.append(p0)
    pos.append(p0.With(chi=0))
    pos.append(p0.With(omega=0))
    pos.append(p0.With(phi=0))
    pos.append(p0.With(chi=90*projtls.degrees))
    pos.append(p0.With(theta=0))

    p0=Position(type='k',
            dx=50,
            hardware=hardware,
            theta=7.23*projtls.degrees,
            omega=-10.00*projtls.degrees,
            kappa=73.00*projtls.degrees,
            phi=100.00*projtls.degrees)

    pos.append(p0)
    pos.append(p0.With(kappa=0))
    pos.append(p0.With(omega=0))
    pos.append(p0.With(phi=0))
    pos.append(p0.With(theta=0))

    def exhaustiveTest(p):
        print('p=',p)
        import traceback
        from goniometer import poscalc
        for type1 in typelist.keys():
            type1=type1.upper()
            if type1 in ['C','B'] and hasattr(p,'theta') and p.theta==0:
                continue
            try:
                p1=eval('p.as_%c()'%type1)
            except(error) as arg:
                if not str(arg)=='Chi high':
                    print('Failed',txt,arg)
                    traceback.print_exc()
                    print()
            except:
                traceback.print_exc()
                print()
            for type2 in typelist.keys():
                type2=type2.upper()
                if type2 in ['C','B'] and hasattr(p,'theta') and p.theta==0:
                    continue
                txt='comparison p.as_%c()==p.as_%c().as_%c()\n'%(type1,type2,type1)
                try:
                    p2=eval('p.as_%c().as_%c()'%(type2,type1))
                except(error) as arg:
                    if not str(arg)=='Chi high':
                        print('Failed1',txt,arg)
                        #traceback.print_exc()
                        #print
                except:
                    print('Failed5',txt)
                    #traceback.print_exc()
                    #print
                try:
                    if poscalc.equal(p1,p2):
                        pass
                        #print('Succeeded',txt)
                    else:
                        print('Failed2',txt)
                        print(p1)
                        print(p2)
                except(error) as arg:
                    if not str(arg)=='Chi high':
                        print('Failed3',txt,arg)
                        #traceback.print_exc()
                        #print
                except:
                    print('Failed4',txt)
                    #traceback.print_exc()
                    #print

    for p in pos:
        exhaustiveTest(p)
