"""Scan parameters and scans.

This module defines classes and functions to calculate, optimize, and perform
scans, and a class to define scan parameters.

The goniometer movement associated with a scan is stored in "scan" objects.  All
other "scan parameters" are stored in "scan parameters" objects (there are
exceptions possible, see the description of the Scan() class for details).

This separation of tasks makes it very easy to change some parameters of the
data collection in one place only (namely in the ScanParams() object), without
touching the data collection strategy. Related, it is possible to generate a
data collection strategy without ever thinking about exactly which frames will
be collected.

A true data collection will thus use a single "scan parameter" object, and one
or multiple "Scan" objects.

"""
#
__version__='$Id: scan.py,v 1.6 2021/02/15 10:22:55 schreurs Exp $'
#
# (C) 1997--2004 Rob W.W. Hooft, Bruker AXS BV
#
import re, os
import numpy

import projtls
from goniometer import position,poscalc
from misc import log,matrix,config

debug=0
minspeed=0.05 # Minimum speed through the Ewald-sphere.

class error(Exception):
    """Exception raised for problems encountered in the scan module.
    """
    pass

def numpat(i):
    """Return a regular expression matching enough '#' characters to fit 'i'

       e.g. if i=4, return '(##*)', for i=1002 return '(#####*)'"""
    s="#*"
    if i>1:
        while i>0:
            s="#"+s
            i=i/10
    return "("+s+")"

def fnpatternvalid(p,scanlimit=99,imagelimit=999):
    """Check the validity of the file name pattern taking limits into account

       arguments:

         **p** -- proposed filenamepattern

         **scanlimit** -- the number of scans it MUST fit

         **imagelimit** -- the number of images per scan it MUST fit
         """
    s1=numpat(scanlimit)
    s2=numpat(imagelimit)
    pre="([^#]*[^0-9#])?"
    if scanlimit<2 and imagelimit<2:
        sep='()'
    else:
        sep="([^#0-9]|[^#0-9][^#]*[^#0-9])"
    aft="([^#0-9][^#]*)?"
    ext="\\.([A-Za-z].*)"
    Rfnpattern=re.compile("^"+pre+s1+sep+s2+aft+ext+"$")
    #print(s1,s2,p)
    return Rfnpattern.match(p)

def newpattern(char):
    """Returns a new file name pattern starting with the given character.
    
       A later character is chosen if the 01/01 file for the given pattern
       already exist"""
    while 1:
        s=char
        for i in range(len(config.hashfilepattern)):
            c=config.hashfilepattern[i]
            if c=='#':
                if config.hashfilepattern[i+1]=='#':
                    s=s+'0'
                else:
                    s=s+'1'
            else:
                s=s+c
        if not os.path.exists(s):
            return char+config.hashfilepattern
        char=chr(ord(char)+1)
    
class ScanParams:
    """Keep the parameters of a scan that are not relevant to simulation.
    """
    def __init__(self,
                 frametime=None,
                 degreetime=None,
                 frameangle=1.0*projtls.degrees,
                 fnpattern='s'+config.hashfilepattern,
                 numimg=2,
                 scanlimit=99,
                 imagelimit=999,
                 firstscan=1,
                 avoidclash=0):
        """Constructor:

           optional parameters:
           
              **frametime** -- Collect data at give number of seconds per frame.

              **degreetime** -- Collect at given number of seconds per degree.
                                If this is given, no frametime should be
                                specified. If none of the two is specified,
                                the default of 10 seconds per frame is used.
                                
              **frameangle** -- Collect frames at given intervals.
              
              **fnpattern** -- The file name pattern to be used.
              
              **numimg** -- number of times each image has to be measured.
                            Normally 2 for dezingering.
                            
              **scanlimit** -- Maximum number of scans that the scan pattern
                               must accommodate.
                               
              **imagelimit** -- The ## in the filename are checked to see
                                if there is space for this many images per scan
                                
              **firstscan** -- number of the first scan to be collected
              
              **avoidclash** -- if 1, repeatedly increment firstscan until file
                                does not exist
        """
        self.degreetime=degreetime
        if frametime:
            if degreetime:
                raise error("frametime and degreetime both given")
            self.frametime=frametime
        elif degreetime:
            self.frametime=frameangle/projtls.degrees*degreetime/numimg
        else:
            self.frametime=10.0
        self.frameangle=frameangle
        self.scanlimit=scanlimit
        self.imagelimit=imagelimit
        match=fnpatternvalid(fnpattern,self.scanlimit,self.imagelimit)
        if not match:
            raise error("Illegal filenamepattern in ScanParams")
        while 1:
            if fnpatternvalid(fnpattern,self.scanlimit*10+9,self.imagelimit):
                self.scanlimit=self.scanlimit*10+9
            else:
                #print("Final scan limit = ",self.scanlimit)
                break
        while 1:
            if fnpatternvalid(fnpattern,self.scanlimit,self.imagelimit*10+9):
                self.imagelimit=self.imagelimit*10+9
            else:
                #print("Final image limit = ",self.imagelimit)
                break
        self.ft1=match.group(1)
        self.fn1=match.group(2)
        self.ft2=match.group(3)
        self.fn2=match.group(4)
        self.ft3=match.group(5) or ""
        self.ext=match.group(6)
        self.numimg=numimg
        self.fnpattern=fnpattern
        if avoidclash:
            while os.path.exists(self.filename(scannr=firstscan,sector=1)):
                firstscan=firstscan+1
        self.firstscan=firstscan
        self.scannr=firstscan-1
        self.ndark=0

    def prefix(self,scannr=None):
        """Calculate the file name prefix.

           optional parameters:
    
             **scannr** -- Number of the scan for which the prefix
                           should be created. If not given, the
                           currently 'active' scan will be used.
                           
           Return value: the prefix string.
        """
        if scannr and scannr>self.scanlimit:
            raise error("Illegal scannumber for scanparameters")
        if scannr is None:
            if len(self.fn1):
                i1=("%0"+str(len(self.fn1))+"d")%self.curscan()
            else:
                i1=""
        else:
            if len(self.fn1):
                i1=("%0"+str(len(self.fn1))+"d")%scannr
            else:
                i1=""
        if self.ft1: 
            return self.ft1+i1+self.ft2
        else:
            return i1+self.ft2
        
    def filename(self,scannr=None,sector=None,extension='kcd'):
        """Calculate a file name.

           optional parameters:

             **scannr** -- Number of the scan for which the prefix
                           should be created. If not given, the
                           currently 'active' scan will be used.

             **sector** -- The image number withing the scan. If not
                           given, the 'hash pattern' for the scan is
                           returned.

             **extension** -- The file name extension.

           Return value: the file name.
        """
        if sector and sector>self.imagelimit:
            raise error("Illegal imagenumber for scanparameters")
        # Remove leading dot from extension
        if extension[0]=='.':
            extension=extension[1:]
        if extension=='kcd':
            extension=self.ext
        if sector is None:
            i2=self.fn2
        else:
            if len(self.fn2):
                i2=("%0"+str(len(self.fn2))+"d")%sector
            else:
                i2=""
        return self.prefix(scannr)+i2+self.ft3+'.'+extension

    def newdarkname(self):
        """Return a new proposed name for a dark current image.

           The returned name will be 'dark##.kcd' with the exposure
           time on the ## marks for the first dark image; for subsequent
           dark images it will be dark##-#.kcd with the last # representing
           a sequence number."""
        self.ndark=self.ndark+1
        if self.ndark>1:
            return "dark%d-%d.%s"%(self.frametime,self.ndark,self.ext)
        else:
            return "dark%d.%s"%(self.frametime,self.ext)

    def resetscan(self):
        """Reset the 'current scan' number.

           This should be called once before each scan set. That call
           is normally handled by the scanset.Perform() method
        """
        self.scannr=self.firstscan-1
        self.ndark=0

    def newscan(self):
        """Progress to the next scan.

           This should be called once before each scan is started. Those
           calls are normally handled by the scanset.Perform() method
        """
        self.scannr=self.scannr+1
        if self.scannr>self.scanlimit:
            raise error("Illegal scannumber for scanparameters")

    def curscan(self):
        """Return the current scan number."""
        return self.scannr

    def __repr__(self):
        """A one-line summary string of the scan parameters."""
        s=""
        if self.frameangle!=None:
            if self.frameangle<0.5*projtls.degrees:
                fmt="%.2f degrees"
            else:
                fmt="%.1f degrees"
            s=s+fmt%(self.frameangle/projtls.degrees)
        if self.degreetime!=None:
            if s:
                s=s+" per frame; "
            s=s+"%.1f seconds per degree; %d iterations."%(self.degreetime,self.numimg)
        elif self.frametime!=None:
            if s:
                s=s+"; "
            s=s+"%i*%.1f seconds per frame."%(self.numimg,self.frametime)
        else:
            if s:
                s=s+" per frame; "
            s=s+"%d iterations."%(self.numimg)
        return s+" Filename %s"%(os.path.basename(self.fnpattern))

    
############################################################################

def varfilter(original,modifier):
    """For modifiable variables, return the modified value.

       parameters:

         **original** -- the original value.

         **modifier** -- the modifier. Should be either 'None' for
                         no modification or ("replace",newvalue)
                         to use the 'newvalue' or ('multiply',factor)
                         to multiply the original value with the given
                         factor.
    """
    if modifier is None:
        return original
    if modifier[0]=='replace':
        return modifier[1]
    if modifier[0]=='multiply':
        return modifier[1]*original
    raise error("Unrecognized modifier %s"%modifier)
        
class AnyScan:
    """Top class for all classes representing a scan."""
    durationmax=5
    
    def __init__(self,p0,p1,
                 anglemargin=config.anglemargin*projtls.degrees,
                 detectorsizemargin=config.detectorsizemargin,
                 step=None,forcenframes=None,frameangle=None,frametime=None):
        """Constructor.

           parameters:
           
              **p0** -- starting position of the scan
              **p1** -- ending position of the scan

           optional parameters:
           
              **anglemargin** -- simulation uses the range start+anglemargin through
                                 end-anglemargin to predict data. This is to account
                                 for reflection size (prevent incomplete partials).
                                 Suggested value: 1--3 degrees.

              **detectorsizemargin** -- Simulation makes detector smaller by 
                                        detectorsizemargin to prevent reflections from
                                        falling of the edge
              **step** -- step size for simulation. If omitted it will be set
                          to a value close to 10 degrees that fills up the 
                          complete scan range. This is only used if the scan
                          is not a rotation that can be predicted using
                          AJMD Impact() calculations.
              **forcenframes** -- Force the number of frames of the hardware
                                  SCAN Default is to follow the scan parameters
                                  coming in, possibly modified by a local
                                  scanangle.  Specifying forcenframes overwrites
                                  all those.
              **frameangle** -- A modifyvariable for the frame angle specified
                               in the scan parameters. Can have three forms::
                                None (default) : just follow the scanparams
                                ('replace',value) : replace scanparams by
                                                    different value.
                                ('multiply',value) : multiply scanparams by
                                                     specified value.
              **frametime** -- A modifyvariable for the frame time specified
                               in the scan parameters.
        """
        self.hardware=p0.hardware
        self.s0=self.hardware.s0
        self.wavelength=self.hardware.wavelength
        self.p0=p0
        self.p1=p1
        self.maxrange=0.0
        if hasattr(p0,'kappa'):
            if self.p0.kappa is not None and self.p1.kappa is not None:
                self.maxrange=max(self.maxrange,abs(self.p0.kappa-self.p1.kappa))
        elif hasattr(p0,'chi'):
            if self.p0.chi is not None and self.p1.chi is not None:
                self.maxrange=max(self.maxrange,abs(self.p0.chi-self.p1.chi))
        if self.p0.dx is not None and self.p1.dx is not None:
            self.maxrange=max(self.maxrange,abs(self.p0.dx-self.p1.dx))
        if self.p0.theta is not None and self.p1.theta is not None:
            self.maxrange=max(self.maxrange,abs(self.p0.theta-self.p1.theta))
        if self.p0.omega is not None and self.p1.omega is not None:
            self.maxrange=max(self.maxrange,abs(self.p0.omega-self.p1.omega))
        if self.p0.phi is not None and self.p1.phi is not None:
            self.maxrange=max(self.maxrange,abs(self.p0.phi-self.p1.phi))
        self.anglemargin=anglemargin
        self.detectorsizemargin=detectorsizemargin
        self.frameangle=frameangle
        self.frametime=frametime
        if forcenframes!=None:
            # Number of frames wanted.
            if self.maxrange==0:
                raise error("Cannot use forcenframes for still image")
            if self.frameangle:
                raise error("Cannot use forcenframes if frameangle specified")
            self.frameangle=('replace',self.maxrange/forcenframes)
        if self.maxrange>0:
            self.startfraction=self.anglemargin/self.maxrange
        else:
            self.startfraction=0
        self.endfraction=1.0-self.startfraction
        if self.endfraction-self.startfraction>0.1:
            if self.maxrange==0:
                self.step=1
            else:
                if step:
                    nstep=int((self.endfraction-self.startfraction)/step+0.5)
                else:
                    # Choose a step size of approximately 7 degrees, but 
                    # take care to fill range completely
                    nstep=int((self.maxrange-2*self.anglemargin)/(7.0*projtls.degrees))+1.0
                if nstep<1:
                    print("Maxrange=",self.maxrange,"margin=",self.anglemargin)
                    print("step=",step)
                    print("fractions,nstep=",self.startfraction,self.endfraction,nstep)
                    raise error("Wrong scan parameters")
                self.step=0.9999999*(self.endfraction-self.startfraction)/nstep
        else:
            self.step=1
            
    def With(self,**kw):
        """Return a modified copy of the scan.

           The original stays intact as if it were an immutable.
           
           possible arguments:

             **p0** -- set the starting position.

             **p1** -- set the ending position.

             **frametime** -- set a new frame time (a modifyvariable!)

             **frameangle** -- set a new frame angle (a modifyvariable!)
        """
        if 'p0' in kw:
            p0=kw['p0']
            del kw['p0']
        else:
            p0=self.p0
        if 'p1' in kw:
            p1=kw['p1']
            del kw['p1']
        else:
            p1=self.p1
        if 'frametime' in kw:
            frametime=kw['frametime']
            del kw['frametime']
        else:
            frametime=self.frametime
        if 'frameangle' in kw:
            frameangle=kw['frameangle']
            del kw['frameangle']
        else:
            frameangle=self.frameangle
        if kw:
            log.BUG("Scan.With: Did not understand kw=%s"%repr(kw),name=__name__)
        return self.__class__(p0,p1,
                              anglemargin=self.anglemargin,
                              detectorsizemargin=self.detectorsizemargin,
                              step=self.step,
                              frameangle=frameangle,
                              frametime=frametime)

    def Storable(self):
        """Return the scan in a form that can be pickled.

           Returns a dictionary that contains all necessary information
           to reconstruct the scan, except for the non-picklable hardware
           reference.
        """
        return {'p0':self.p0.Storable(),
                'p1':self.p1.Storable(),
                'anglemargin':self.anglemargin,
                'detectorsizemargin':self.detectorsizemargin,
                'frameangle':self.frameangle,
                'frametime':self.frametime,
                'step':self.step}

    def _fracpos(self,fraction):
        """Return a position that is 'fraction' along the way of the scan"""
        if fraction<-0.0000001 or fraction>1.0000001:
            raise IndexError("fracpos out of range: %.10f"%fraction)
        return self.p0.towards(self.p1,fraction)
        
    def realstartposition(self,offset=0.0):
        """Return a position that corresponds to the start of the scan

           optional parameters:

             **offset** -- the rotational offset (in radians) along the
                           fastest rotating axis.
        """
        if self.maxrange==0:
            return self._fracpos(0.0)
        else:
            return self._fracpos(offset/self.maxrange)

    def __getitem__(self,i):
        """Return the position corresponding to the i-th simulation step.
        """
        fraction=self.startfraction+i*self.step
        if fraction>self.endfraction:
            raise IndexError("End of loop?")
        return self._fracpos(self.startfraction+i*self.step)

    def _frameangle(self,scanparams):
        """Return the actual frame angle.

           The actual value is determined from the frame angle specified in the
           scan parameters, and the modifyvariable stored at construction time.
        """
        return varfilter(scanparams.frameangle,self.frameangle)

    def _frametime(self,scanparams):
        """Return the actual frame time.

           The actual value is determined from the frame time specified in the
           scan parameters, and the modifyvariable stored at construction time.
        """
        if scanparams.degreetime:
            ftime=self._frameangle(scanparams)*scanparams.degreetime/projtls.degrees/scanparams.numimg
        else:
            ftime=scanparams.frametime
        return varfilter(ftime,self.frametime)

    def _ad(self,scanparams):
        """Return the number of additional images next to the scan.

           Normally returns 1 if an autodark image will be made.
        """
        if config.autodark and self.ScanImages(scanparams)>1:
            return 1
        return 0
        
    def NumImages(self,scanparams):
        """Return the total number of images that will be made by this scan.

           That includes the dark image (if any).
        """
        #print("DBG> NumImages=",self.ScanImages(scanparams)+self._ad(scanparams))
        return self.ScanImages(scanparams)+self._ad(scanparams)
        
    def ScanImages(self,scanparams):
        """Return the numbrt of images in the scan.

           This does not include any dark image.
        """
        if self._frameangle(scanparams)==0.0 and self.maxrange==0:
            # Still image
            return 1
        else:
            #print("DBG> maxrange=",self.maxrange)
            #print("DBG> frameangle=",self._frameangle(scanparams))
            return int(abs((self.maxrange*1.0000001)/self._frameangle(scanparams)))

    def ImageTime(self,scanparams):
        """Return the time that will be used for a single image."""
        return self.NumImages(scanparams)*scanparams.numimg*self._frametime(scanparams)
        
    def ExpectedRunTime(self,scanparams):
        """Part of the Operation() protocol. How long will this take?"""
        ni=self.NumImages(scanparams)
        t1=ni*(self.hardware.overheadtime)*scanparams.numimg # Lost time
        if scanparams.degreetime is not None:
            t2=self.ScanLength()/projtls.degrees
        else:
            t2=ni*1.0*scanparams.numimg  # Extra time per second of exposure.
        if self.frametime:
            if self.frametime[0]=='replace':
                t1=t1+t2*self.frametime[1]
                t2=0
            elif self.frametime[0]=='multiply':
                t2=t2*self.frametime[1]
        #print("DBG> t1,t2=",t1,t2)
        return (t1,t2)
    
    def ScanLength(self):
        """Part of the Operation() protocol.

           Calculate the length of the scan (in radians)
        """
        return self.maxrange

    def frametimerepr(self):
        """A zero or one-line string showing the frametime modifier"""
        if self.frametime and self.frametime!=('multiply',1.0):
            if self.frametime[0]=='multiply':
                if self.frametime[1]<1.0:
                    f="%5.3f"%self.frametime[1]
                else:
                    f="%5.1f"%self.frametime[1]
                return "multiply measurement time by "+f
            elif self.frametime[0]=='replace':
                return "measure for %.1f seconds"%self.frametime[1]
            else:
                return "Illegal frametime %s"%self.frametime
        return ""

    def frameanglerepr(self):
        """A zero or one-line string showing the frameangle modifier"""
        if self.frameangle and self.frameangle!=('multiply',1.0):
            if self.frameangle[0]=='multiply':
                return "multiply frame angle by %4.2f"%self.frameangle[1]
            elif self.frameangle[0]=='replace':
                return "frame angle %.2f degrees"%(self.frameangle[1]/projtls.degrees)
            else:
                return "Illegal frame angle %s"%self.frameangle
        return ""

    def modrepr(self):
        """A string representing the variable modifiers"""
        t1=self.frametimerepr()
        t2=self.frameanglerepr()
        if t1 and t2:
            return " \n%s; %s"%(t1,t2)
        elif t1:
            return " \n%s"%(t1)
        elif t2:
            return " \n%s"%(t2)
        else:
            return ""

    def __repr__(self):
        """A total string representation of the scan."""
        if poscalc.equal(self.p0,self.p1):
            return "<%s: still at %s%s>"%(
                self.__class__.__name__,
                self.p0,
                self.modrepr()
                )
        else:
            return "<%s: from %s\n through %s%s>"%(
                self.__class__.__name__,
                self.p0,
                self.p1,
                self.modrepr()
                )

    def allowed(self):
        """Check whether the scan can be performed on the hardware."""
        return self.hardware.RouteAllowed(self.p0,self.p1)

    def copy(self):
        """Make a copy of the scan."""
        import copy
        return copy.copy(self)
    
class OneAxisScan(AnyScan):
    """A 'Rotation' scan in which one motor of the Kappa goniostat moves.

       Do not instantiate this class, but one of the subclasses, eg
       'PhiScan' or 'OmegaScan' instead.
    """
    def __init__(self,start,end,p,
                 anglemargin=config.anglemargin*projtls.degrees,
                 detectorsizemargin=config.detectorsizemargin,
                 step=None,frameangle=None,frametime=None,forcenframes=None):
        """Constructor.

           parameters:
           
              **start** -- starting angle of the physical scan

              **end**   -- ending angle of the physical scan

              **p**     --  position of all constant angles

           optional parameters:
           
              See AnyScan() constructor.
        """
        self.argstart=start # attribute used by makebox!
        self.argend=end # attribute used by makebox!
        self.argp=p
        p0=self.makep0(self.argp,start)
        p1=self.makep1(self.argp,end)
        AnyScan.__init__(self,
                         p0=p0,
                         p1=p1,
                         anglemargin=anglemargin,
                         detectorsizemargin=detectorsizemargin,
                         step=step,
                         forcenframes=forcenframes,
                         frameangle=frameangle,
                         frametime=frametime)


    def With(self,**kw):
        """Return a modified copy of the scan.

           The original stays intact as if it were an immutable.
           
           possible arguments:

             **p0** -- set the starting position.

             **p1** -- set the ending position.

             **start** -- set the starting value of the rotation angle.

             **end** -- set the end value of the rotation angle.

             **frametime** -- set a new frame time (a modifyvariable!)

             **frameangle** -- set a new frame angle (a modifyvariable!)
        """
        p=self.p0
        if 'p0' in kw:
            p0=kw['p0']
            del kw['p0']
            p=p0
        else:
            p0=self.p0
        if 'p1' in kw:
            p1=kw['p1']
            del kw['p1']
            p=p1
        else:
            p1=self.p1
        if 'start' in kw:
            p0=self.makep0(p0,kw['start'])
            del kw['start']
        if 'end' in 'end':
            p1=self.makep1(p1,kw['end'])
            del kw['end']
        if 'frametime' in kw:
            frametime=kw['frametime']
            del kw['frametime']
        else:
            frametime=self.frametime
        if 'frameangle' in kw:
            frameangle=kw['frameangle']
            del kw['frameangle']
        else:
            frameangle=self.frameangle
        if kw:
            log.BUG("Scan.With: Did not understand kw=%s"%kw,name=__name__)
        return self.__class__(self.rotval(p0),self.rotval(p1),p,
                              anglemargin=self.anglemargin,
                              detectorsizemargin=self.detectorsizemargin,
                              step=self.step,
                              frameangle=frameangle,
                              frametime=frametime)

    def StartAngle(self):
        """The starting value of the rotating angle"""
        return self.rotval(self.p0)
    
    def EndAngle(self):
        """The ending value of the rotation angle"""
        return self.rotval(self.p1)
    
    def Storable(self):
        """Return the scan in a form that can be pickled.

           Returns a dictionary that contains all necessary information
           to reconstruct the scan, except for the non-picklable hardware
           reference.
        """
        return {'angle':self.angle,
                'start':self.argstart,
                'end':self.argend,
                'p':self.argp.Storable(),
                'anglemargin':self.anglemargin,
                'detectorsizemargin':self.detectorsizemargin,
                'frameangle':self.frameangle,
                'frametime':self.frametime,
                'step':self.step}

class PhiScan(OneAxisScan):
    """A "OneAxisScan" that will rotate around the phi angle.

       a PhiScan should in general only be made while the Phi axis is
       perpendicular to the beam. Generally only one phi scan is made at zero
       kappa (or a low nonzero value of kappa if the low temp device is active),
       and the rest of the data is collected in omega scans at non-zero values
       of kappa.
 
       See the description of the superclass OneAxisScan.
    """
    angle="phi"
    def makep0(self,p,start):
        """Return a Position() for the beginning of the scan"""
        return p.With(phi=start)
    
    def makep1(self,p,end):
        """Return a Position() for the end of the scan"""
        return p.With(phi=end)

    def rotval(self,p):
        """Return the rotation value for p is a Position() in the scan"""
        return p.phi

    def RotationAxis(self):
        """The position of the rotation axis in space."""
        return matrix.matvec(self.p0.goniometermatrix(),numpy.array([0,0,1]))

    def __repr__(self):
        """Short summary string describing the scan"""
        if config.scanrepr==0:
            return "<Phi scan:%6.1f through %6.1f; base=%s%s>"%(
                self.p0.phi/projtls.degrees,
                self.p1.phi/projtls.degrees,
                self.p0.With(phi=None),
                self.modrepr()
                )
        elif config.scanrepr==1:
            return "<Phi   scan: len=%6.1f; start=%s%s>"%(
                (self.p1.phi-self.p0.phi)/projtls.degrees,
                self.p0,
                self.modrepr()
                )
        else:
            return "<Phi   scan: start=%6.1f; len=%6.1f; end=%6.1f at %s%s>"%(
                (self.p0.phi)/projtls.degrees,
                (self.p1.phi-self.p0.phi)/projtls.degrees,
                (self.p1.phi)/projtls.degrees,
                self.p0.With(phi=None),
                self.modrepr()
                )
            
class OmegaScan(OneAxisScan):
    """A "OneAxisScan" that will rotate around the omega angle.

       an OmegaScan has the advantage that the omega axis is always
       perpendicular to the beam and parallel to the detector. Hence
       omega scans are the safest bet to collect data in an efficient
       way. The problem is that omega scans can in general not be very
       long, especially at low values of dx and at high absolute
       values of kappa.

       See the description of the superclass OneAxisScan.
    """
    angle="omega"
    def makep0(self,p,start):
        """Return a Position() for the beginning of the scan"""
        return p.With(omega=start)
    
    def makep1(self,p,end):
        """Return a Position() for the end of the scan"""
        return p.With(omega=end)

    def rotval(self,p):
        """Return the rotation value for p is a Position() in the scan"""
        return p.omega

    def __repr__(self):
        """Short summary string describing the scan"""
        if config.scanrepr==0:
            return "<Omega scan:%6.1f through %6.1f; base=%s%s>"%(
                self.p0.omega/projtls.degrees,
                self.p1.omega/projtls.degrees,
                self.p0.With(omega=None),
                self.modrepr()
                )
        elif config.scanrepr==1:
            return "<Omega scan: len=%6.1f; start=%s%s>"%(
                (self.p1.omega-self.p0.omega)/projtls.degrees,
                self.p0,
                self.modrepr()
                )
        else:
            return "<Omega scan: start=%6.1f; len=%6.1f; end=%6.1f at %s%s>"%(
                (self.p0.omega)/projtls.degrees,
                (self.p1.omega-self.p0.omega)/projtls.degrees,
                (self.p1.omega)/projtls.degrees,
                self.p0.With(omega=None),
                self.modrepr()
                )

    def RotationAxis(self):
        """The position of the rotation axis in space.

           Always vertical.
        """
        return numpy.array([0,0,1])
                             
def MaximalOmegaScan(p,
                     anglemargin=config.anglemargin*projtls.degrees,
                     detectorsizemargin=config.detectorsizemargin):
    """Return an OmegaScan() that uses as much space as possible around 'p'.

       This queries the hardware for allowed space.
    """
    hwrange=p.orange()
    if not hwrange:
        return None # Next try. We cannot reach this position.
    minang=hwrange[0]
    maxang=hwrange[1]
    if not (minang+anglemargin<p.omega<maxang-anglemargin):
        return None # Too small range. We cannot make a serious scan here.
    start=minang+anglemargin
    end=maxang-anglemargin
    return OmegaScan(start=start,
                     end=end,
                     p=p,
                     anglemargin=anglemargin,
                     detectorsizemargin=detectorsizemargin)

def SymmetricOmegaScan(p,width,
                       anglemargin=config.anglemargin*projtls.degrees,
                       detectorsizemargin=config.detectorsizemargin):
    """Return an OmegaScan() that starts width/2 before p, and ends width later.

       this queries the hardware for allowed space, and will deviate from
       centering on 'p' if there is not enough space on one of the two sides.
    """
    hwrange=p.orange()
    if not hwrange:
        return None
    minang=hwrange[0]
    maxang=hwrange[1]
    #print("minang,maxang=",hwrange[0]/projtls.degrees,hwrange[1]/projtls.degrees)
    start=p.omega-width/2.0
    if start<minang+anglemargin:
        start=minang
    end=start+width
    if end>maxang-anglemargin:
        return None
    #print("start,end=",start/projtls.degrees,end/projtls.degrees)
    return OmegaScan(start=start,
                     end=end,
                     p=p,
                     anglemargin=anglemargin,
                     detectorsizemargin=detectorsizemargin)

def ScanFromStorage(hardware,state):
    """Reconstruct a scan from the dictionary 'state'.

       Automatically recognizes the type of scan from the state dictionary, and
       constructs the scan applied to the given hardware connection. The
       construction verifies that the scan can be performed on the hardware. If
       only small adjustments are required to the endpoints, these are
       automatically applied. If larger adjustments would be required, 'None' is
       returned.
    """
    if 'p1' in state:
        # AnyScan
        p0=position.PositionFromStorage(hardware,state['p0'])
        p1=position.PositionFromStorage(hardware,state['p1'])
        if hardware.RouteAllowed(p0,p1):
            return AnyScan(p0=p0,p1=p1,
                           anglemargin=state['anglemargin'],
                           detectorsizemargin=state['detectorsizemargin'],
                           step=state['step'],
                           frameangle=state.get('frameangle'),
                           frametime=state.get('frametime'))
        else:
            print("Warning: Any scan from save file ignored. Collision.")
            return None
    # Phi or Omega scan
    p=position.PositionFromStorage(hardware,state['p'])
    if state['angle']=='phi':
        aclass=PhiScan
        orange=p.orange()
        if not orange:
            print("Warning: Phi scan from save file ignored. Impossible position.")
            return None
    elif state['angle']=='omega':
        aclass=OmegaScan
        orange=p.With(omega=(state['start']+state['end'])/2.0).orange()
        if orange:
            if not p.With(omega=state['start']+0.0005*projtls.degrees).allowed():
                print("Warning: Start value of omega scan from save file adjusted.")
                state['start']=orange[0]
            if not p.With(omega=state['end']-0.0005*projtls.degrees).allowed():
                print("Warning: End value of omega scan from save file adjusted.")
                state['end']=orange[1]
        or0=p.With(omega=state['start']+0.0005*projtls.degrees).orange()
        or1=p.With(omega=state['end']-0.0005*projtls.degrees).orange()
        if or0 != or1:
            print("Warning: Scan from save file ignored. Impossible position(s).")
            print(state)
            print(or0, or1)
            return None
    else:
        raise error('Unknown scan type from Storage. Reloading newer version?')
    return aclass(start=state['start'],
                  end=state['end'],
                  p=p,
                  anglemargin=state['anglemargin'],
                  detectorsizemargin=state['detectorsizemargin'],
                  step=state['step'],
                  frameangle=state.get('frameangle'),
                  frametime=state.get('frametime'))

def ScanFromKcdHeader(h,hardware=None,extendfactor=1.0,hend=None,mosaicity=0.0):
    """Reconstruct a Scan() object from an image file header.

       parameters:

         **h** -- image header.

       optional parameters:
       
         **extendfactor** -- make the scan this many times longer than the
                             image header **h** indicates.

         **hend** -- the image header of the last image in the scan. This
                     is incompatible with the **extendfactor** parameter, and is
                     preferable when a scan should be constructed that
                     corresponds to a complete set of images.

         **mosaicity** -- an extra angle to be added before the beginning and
                          after the end of the scan to allow for mosaicity of
                          the crystal.
    """
    if hardware is None:
        from hwdriver import dummydrv
        hardware=dummydrv.DummyGon(header=h)
    basepos=poscalc.fromheader(h,fraction=0.0,hardware=hardware)
    e1pos=poscalc.fromheader(h,fraction=1.0,hardware=hardware)
    dom=abs(e1pos.omega-basepos.omega)
    dph=abs(e1pos.phi-basepos.phi)
    if dom<0.001*projtls.degrees and dph>0.001*projtls.degrees:
        if e1pos.phi<basepos.phi:
            angle='phi'
            direction=-1
        else:
            angle='phi'
            direction=1
    elif dph<0.001*projtls.degrees and dom>0.001*projtls.degrees:
        if e1pos.omega<basepos.omega:
            angle='omega'
            direction=-1
        else:
            angle='omega'
            direction=1
    else:
        angle='unknown'
        direction=0
    if hend:
        hendpos=poscalc.fromheader(hend,fraction=1.0,hardware=hardware)
        if angle=='omega' and direction<0:
            if hendpos.omega>basepos.omega:
                hendpos=hendpos.With(omega=hendpos.omega-360*projtls.degrees)
                print("DBG> Corrected endpos, omega turns negative but ended higher")
        elif angle=='omega' and direction>0:
            if hendpos.omega<basepos.omega:
                hendpos=hendpos.With(omega=hendpos.omega+360*projtls.degrees)
                print("DBG> Corrected endpos, omega turns positive but ended lower")
        elif angle=='phi' and direction<0:
            if hendpos.phi>basepos.phi:
                hendpos=hendpos.With(phi=hendpos.phi-360*projtls.degrees)
                print("DBG> Corrected endpos, phi turns negative but ended higher")
        elif angle=='phi' and direction>0:
            if hendpos.phi<basepos.phi:
                hendpos=hendpos.With(phi=hendpos.phi+360*projtls.degrees)
                print("DBG> Corrected endpos, phi turns positive but ended lower")
        else:
            pass
            # Can not correct, not a simple omega or phi scan.
    else:
        hendpos=None
    if extendfactor:
        eendpos=poscalc.fromheader(h,fraction=extendfactor,hardware=hardware)
    else:
        eendpos=None
    if hendpos:
        endpos=hendpos
    elif eendpos:
        endpos=eendpos
    if not hasattr(h,'scanangle'):
        dphi=abs(endpos.phi-basepos.phi)
        domega=abs(endpos.omega-basepos.omega)
        maxrot=max(dphi,domega)
        if hasattr(basepos,'chi'):
            chiequal=abs(basepos.chi-endpos.chi)<maxrot/1000
        else:
            chiequal=abs(basepos.kappa-endpos.kappa)<maxrot/1000
        if abs(basepos.dx-endpos.dx)<maxrot/100 and chiequal and abs(basepos.theta-endpos.theta)<maxrot/1000:
            #print("DBG> ",dphi,domega)
            if dphi<domega/1000:
                #print("DBG> OMEGASCAN")
                return OmegaScan(start=basepos.omega-0.5*mosaicity,end=endpos.omega+0.5*mosaicity,p=basepos,anglemargin=0)
            elif domega<dphi/1000:
                #print("DBG> PHISCAN")
                return PhiScan(start=basepos.phi-0.5*mosaicity,end=endpos.phi+0.5*mosaicity,p=basepos,anglemargin=0)
        if mosaicity:
            print("Ignoring mosaicity in ScanFromKcdHeader for non-rotation scan")
        #print("DBG> ANYSCAN")
        return AnyScan(p0=basepos,p1=endpos,anglemargin=0)
    elif h.scanangle=='Phi':
        return PhiScan(start=basepos.phi-0.5*mosaicity,end=endpos.phi+0.5*mosaicity,p=basepos,anglemargin=0)
    elif h.scanangle=='Omega':
        return OmegaScan(start=basepos.omega-0.5*mosaicity,end=endpos.omega+0.5*mosaicity,p=basepos,anglemargin=0)
    else:
        raise error("Cannot represent scan from KCD header")
