from misc import evpyprint
newline=evpyprint.newline
nonewline=evpyprint.nonewline
#
__version__='$Id: rmat.py,v 1.10 2020/08/30 10:51:39 schreurs Exp $'
#
# Read (and understand) dirax .rmat files.
#
# (C) 1998--2004 Rob W. W. Hooft, Bruker AXS BV
#
import re
import numpy

import projtls
from goniometer import scan
from misc import log

def tonew(pointgroup):
    upg=pointgroup.upper()
    if upg in ['MB','M']:
        return '2/m'
    elif upg in ['MC']:
        return '112/m'
    elif upg in ['MA']:
        return '2/m11'
    elif upg in ['O']:
        return 'mmm'
    elif upg in ['T','TL']:
        return '4/m'
    elif upg in ['TH']:
        return '4/mmm'
    elif upg in ['H','HL']:
        return '6/m'
    elif upg in ['HH']:
        return '6/mmm'
    elif upg in ['C','CL']:
        return 'm3'
    elif upg in ['CH']:
        return 'm3m'
    else:
        return pointgroup.lower()

pointgroup2system={'-1':'triclinic',
                   '1':'triclinic',
                   '2':'monoclinic',
                   'm':'monoclinic',
                   '2/m':'monoclinic',
                   '112':'monoclinic',
                   '11m':'monoclinic',
                   '112/m':'monoclinic',
                   '211':'monoclinic',
                   'm11':'monoclinic',
                   '2/m11':'monoclinic',
                   'mmm':'orthorhombic',
                   'mm2':'orthorhombic',
                   'm2m':'orthorhombic',
                   '2mm':'orthorhombic',
                   '222':'orthorhombic',
                   '4':'tetragonal',
                   '-4':'tetragonal',
                   '4/m':'tetragonal',
                   '422':'tetragonal',
                   '4mm':'tetragonal',
                   '-42m':'tetragonal',
                   '4/mmm':'tetragonal',
                   '3':'trigonalh',
                   '-3':'trigonalh',
                   '32':'trigonalh',
                   '3m1':'trigonalh',
                   '-3m1':'trigonalh',
                   '31m':'trigonalh',
                   '-31m':'trigonalh',
                   '6':'hexagonal',
                   '-6':'hexagonal',
                   '6/m':'hexagonal',
                   '622':'hexagonal',
                   '6mm':'hexagonal',
                   '-6m2':'hexagonal',
                   '6/mmm':'hexagonal',
                   '23':'cubic',
                   'm3':'cubic',
                   '432':'cubic',
                   '-43m':'cubic',
                   'm3m':'cubic'}

Rnumbers=re.compile('# +H +(\d+) +n (\d+)')

class Rmat:
    """Read Dirax .rmat

       Public Methods:
            GetOrientationMatrix
    """
    def __init__(self,rmatfilename=None,rmatstr=None,ignoretmat=0):
        """Constructor

           parameters:
               rmatfilename : name of .rmat file
        """
        if rmatfilename:
            f=projtls.myopen(rmatfilename,'r')
            rmatstr=f.read()
            f.close()
            if rmatstr[-1]!='\n':
                log.WARNING('Incomplete last line on %s'%rmatfilename)
        self.lines=rmatstr.split('\n')
        self.lineno=-1
        self.q=[]
        self.qvc=[]
        self._readrmat(ignoretmat)
        
    def nextline(self):
        self.lineno=self.lineno+1
        try:
            return self.lines[self.lineno]
        except IndexError:
            return ''
    
    def _readrmat(self,ignoretmat=0):
        # Skip leading comment lines
        while 1:
            line=self.nextline()
            if not line:
                break
            if line[0]=='#':
                # See if this comment line gives some statistics.
                m=Rnumbers.match(line)
                if m:
                    self.numh=int(m.group(1))
                    self.numn=int(m.group(2))
                # Skip comment lines
                continue
            word=line.split()
            if not word:
                continue
            cmd=word[0].lower()
            if cmd=='rmat':
                if len(word)==1:
                    self.bravais='P'
                else:
                    self.bravais=word[1].upper()
                    if len(self.bravais)>1:
                        self.bravais='P'
                # Interpret the next lines as the matrix
                (a11,a12,a13)=projtls.stringtofloats(self.nextline())
                (a21,a22,a23)=projtls.stringtofloats(self.nextline())
                (a31,a32,a33)=projtls.stringtofloats(self.nextline())
                # Change the axis system to our standard.
                self.orientationmatrix=numpy.transpose(
                    [[a11,a12,a13],[a21,a22,a23],[a31,a32,a33]])
                self.tmatrix=None
            elif cmd=='qvec':
                if self.qvc:
                    log.ERROR("QVEC encountered in rmat after QVC. Rejecting QVCs so far")
                    self.qvc=[]
                try:
                    if len(word) == 4:
                        log.WARNING("assuming qvector order 0")
                        word.append('0')
                    self.q.append( (float(word[1]),float(word[2]),float(word[3]),
                                     int(word[4])))
                except (IndexError,ValueError):
                    log.ERROR("Invalid Q vector in RMAT. Ignored.")
            elif cmd=='qvc':
                if len(word)!=len(self.q)+1:
                    log.ERROR("Ignored QVC with %d instead of %d Qvectors"%(len(word)-1,len(self.q)))
                else:
                    self.qvc.append(map(int,word[1:]))
            elif cmd=='tmat':
                # Interpret the next lines as the matrix
                (t11,t12,t13)=projtls.stringtofloats(self.nextline())
                (t21,t22,t23)=projtls.stringtofloats(self.nextline())
                (t31,t32,t33)=projtls.stringtofloats(self.nextline())

                if not ignoretmat:
                    self.tmatrix=numpy.transpose(
                        [[t11,t12,t13],[t21,t22,t23],[t31,t32,t33]])
                    if len(word)>1:
                        self.transformedbravais=word[1]
                    if len(word)>2:
                        self.pointgroup=tonew(word[2])
            elif cmd=='sigma' or cmd=='sigmacell':
                self.sigma=list(map(float,word[1:]))
                self.sigma[3]=self.sigma[3]*projtls.degrees
                self.sigma[4]=self.sigma[4]*projtls.degrees
                self.sigma[5]=self.sigma[5]*projtls.degrees
            elif cmd=='face':
                pass # Ignore command lines for absorption faces.
            elif cmd=='cell':
                pass # Ignore command lines for unit cell parameters.
            elif cmd=='qvectype':
                pass # Ignore
            else:
                log.WARNING("Command %s in .rmat file ignored"%cmd)
                
    def CrystalSystem(self):
        if hasattr(self,'pointgroup'):
            try:
                return pointgroup2system[self.pointgroup]
            except KeyError:
                log.WARNING("Unrecognized point group %s"%self.pointgroup)
                return None
        else:
            return None

    def PointGroup(self):
        if hasattr(self,'pointgroup'):
            return self.pointgroup
        else:
            return None
        
    def Sigma(self):
        if hasattr(self,'sigma'):
            return self.sigma
        else:
            return None

    def GetOrientationMatrix(self,hardware=None):
        """Calculate orientation matrix of the crystal in CAD4 coordinates

           optional parameters: 
               hardware:  hardware object, ignored

           return value:
               a 3x3 orientation matrix (row matrix of a* b* c*)
        """
        return self.orientationmatrix

    def GetTransformationMatrix(self):
        #if self.tmatrix:
            #newline("Using RMAT supplied transformation matrix")
            #newline(self.tmatrix)
        return self.tmatrix
    
    def Bravaistype(self):
        return self.bravais

    def TransformedBravaistype(self):
        if hasattr(self,'transformedbravais'):
            return self.transformedbravais

def RmatCell(rmatfilename=None,rmatstr=None,hardware=None,standardize=1,
             axcrit=None,transformtrigonal=1,bravaistype=None,
             pointgroup=None,wavelength=None,ignoretmat=0):
    if pointgroup=='Default':
        pointgroup=None
    if axcrit is None:
        from experiment import hints
        axcrit=hints.axcrit()
    if rmatfilename:
        X=Rmat(rmatfilename,ignoretmat=ignoretmat)
    elif rmatstr:
        X=Rmat(rmatstr=rmatstr,ignoretmat=ignoretmat)
    else:
        raise ValueError("Need either rmatfilename or rmatstr")

    from experiment import crystal
    try:
        om=X.GetOrientationMatrix(hardware)
    except AttributeError:
        raise ValueError("No cell in file")
    if pointgroup:
        if X.PointGroup() and pointgroup!=X.PointGroup():
            log.WARNING("Specified point group %s differs from RMAT file %s"%(pointgroup,X.PointGroup()))
    else:
        pointgroup=X.PointGroup()
    if bravaistype:
        if X.Bravaistype() and bravaistype!=X.Bravaistype():
            log.WARNING("Bravais type argument %s differs from RMAT file %s"%(bravaistype,X.Bravaistype()))
    else:
        bravaistype=X.Bravaistype()
    xtra={'filename':rmatfilename}
    if hasattr(X,'numh'):
        xtra['numh']=X.numh
    if hasattr(X,'numn'):
        xtra['numn']=X.numn
    c=crystal.Cell(orientationmatrix=om,
                   wavelength=wavelength,
                   hardware=hardware,
                   axcrit=axcrit,
                   standardize=standardize,
                   bravaistype=bravaistype,
                   transformedbravaistype=X.TransformedBravaistype(),
                   system=X.CrystalSystem(),
                   pointgroup=pointgroup,
                   transformtrigonal=transformtrigonal,
                   transformationmatrix=X.GetTransformationMatrix(),
                   xtra=xtra)
    s=X.Sigma()
    if s:
        if numpy.any(X.GetTransformationMatrix()):
            valid=1
            #newline("DBG> Sigma's valid: Transformation given")
        else:
            t=numpy.absolute(c.input2conventionaltransformation())
            t=t-numpy.identity(3)
#            a=Numeric.add.reduce(Numeric.ravel(t))
            a=numpy.sum(t)
            if a<0.1:
                #newline("DBG> Sigma's valid: Cells identical")
                valid=1
            else:
                log.NOTE("Sigma's become invalid: Cell is transformed")
                valid=0
        if valid:
            c.SetSigma(s)
    if X.q:
        if numpy.any(X.GetTransformationMatrix()):
            # Qvectors valid: Transformation given
            qq=X.q
        else:
            t=c.input2conventionaltransformation()
            t2=numpy.absolute(t)-numpy.identity(3)
#            a=Numeric.add.reduce(Numeric.ravel(t2))
            a=numpy.sum(t2)
            if a<0.1:
                # Qvectors valid: Cells identical"
                qq=X.q
            else:
                # Qvectors become invalid: Cell is transformed
                from misc import matrix
                qq=[]
                for qvec in X.q:
                    qout=tuple(matrix.matvec(t,qvec[:3]))+(qvec[3],)
                    newline(qvec,qout)
                    qq.append(qout)
        for qvec in qq:
            c.AddQvec(qvec)
        if X.qvc:
            c.SetQvc(X.qvc)
    return c

if __name__=="__main__":
    import os,sys
    from experiment import hints
    from misc import args

    xfilename=hints.xfilename()
    outmatname=None
    if xfilename and xfilename[-2:]=='.x':
        xfilename=projtls.youngest('*.rmat')
    standardize=1
    transformtrigonal=1
    transformed=0
    pointgroup=None
    axcrit=hints.axcrit()
    showpg=0
    ignoretmat=0
    for word in args.ArgLoop():
        if word=='nostandardize':
            standardize=0
        elif word=='notransformtrigonal':
            transformtrigonal=0
        elif word=='transformed':
            transformed=1
        elif word=='showpg':
            showpg=1
        elif word=='ignoretmat':
            ignoretmat=1
        elif word.sets('axcrit'):
            axcrit=word.angleval()
            hints.axcrit(axcrit)
        elif word[-5:]=='.rmat' and os.path.exists(word.filename()):
            xfilename=word.filename()
        elif word.sets('write'):
            outmatname=word.stringval()
        elif word.sets('pointgroup'):
            pointgroup=word.stringval()
        elif word=="-h":
            newline('Usage: rmatrix nostandardize notransformtrigonal transformed showpg axcrit=angle write=filename pointgroup=pointgroupname')
        else:
            word.unknown()
    if xfilename is None:
        log.ERROR("No input cell found, please specify .rmat file")
        sys.exit(1)
            
    c=RmatCell(xfilename,axcrit=axcrit,standardize=standardize,
               transformtrigonal=transformtrigonal,
               pointgroup=pointgroup,ignoretmat=ignoretmat)
    newline(c)
    if outmatname:
        f=projtls.myopen(outmatname,'w')
        if transformed:
            f.write(c.transformedrmatstr())
            newline("Transformed matrix written to %s"%outmatname)
        else:
            f.write(c.rmatstr())
            newline("Standardized matrix written to %s"%outmatname)
        f.close()
    else:
        if transformed:
            nonewline(c.transformedrmatstr())
        else:
            nonewline(c.rmatstr())
    if showpg:
        newline("Valid point groups: ",c.ValidPointgroups())

