#!/usr/bin/env python3
from misc import evpyprint
newline=evpyprint.newline
nonewline=evpyprint.nonewline
import os
import sys
import numpy
import scipy.stats
import projtls
import argparse
import re
from interface import rmat
from misc import plural
_prog_usage = '%(prog)s [options]'
_prog_version = 'loglog Version 1.3'
_prog_description = 'scatterplots of two reflection files'
_prog_author = 'Toine Schreurs'

raw_input = input

DEBUGMAXREFLS = 100
EXTRAUPPER = 1.0e-6
DEGREES = numpy.pi/180.0
CHANGESIGMA = True
INCH=2.54
GAP = 0.01
SUBWINDOWWIDTH = 0.15

class ActionNoYes(argparse.Action):
    def __init__(self, option_strings, dest, default=None, required=False, help=None):

        if default is None:
            raise ValueError('You must provide a default with Yes/No action')
        if len(option_strings)!=1:
            raise ValueError('Only single argument is allowed with YesNo action')
        opt = option_strings[0]
        if not opt.startswith('--'):
            raise ValueError('Yes/No arguments must be prefixed with --')

        opt = opt[2:]
        opts = ['--' + opt, '--no-' + opt]
        super(ActionNoYes, self).__init__(opts, dest, nargs=0, const=None,
                                          default=default, required=required, help=help)
    def __call__(self, parser, namespace, values, option_strings=None):
        if option_strings.startswith('--no-'):
            setattr(namespace, self.dest, False)
        else:
            setattr(namespace, self.dest, True)


def absvec(v):
    """Determine length of single 3D vector or each out of an array of vectors"""
    if v.shape[-1]!=3:
        raise ValueError("Invalid shape for v in vectorlength")
    return numpy.sqrt(v[...,0]*v[...,0]+v[...,1]*v[...,1]+v[...,2]*v[...,2])

def gen117(n):

    def gen116(p):
        p0 = -0.322232431088
        p1 = -1.0
        p2 = -0.342242088547
        p3 = -0.204231210245E-1
        p4 = -0.453642210148E-4
        q0 = 0.993484626060E-1
        q1 = 0.588581570495
        q2 = 0.531103462366
        q3 = 0.103537752850
        q4 = 0.38560700634E-2

        if p == 0.5:
            ppf = .0
        else:
            r = p
            if p > 0.5:
                r = 1.0-r
            t = numpy.sqrt(-2.0*numpy.log(r))
            anum = ((((t*p4+p3)*t+p2)*t+p1)*t+p0)
            aden = ((((t*q4+q3)*t+q2)*t+q1)*t+q0)
            ppf = t + (anum/aden)
            if p < 0.5:
                ppf = -ppf
        return ppf

    a = numpy.zeros(n)
    f = 0.5**(1.0/n)
    a[0] = gen116( 1.0-f )
    a[n-1] = gen116( f ) 
    if n%2 != 0:
        nhalf = n//2 
        a[nhalf] = gen116( 0.5 )
    gam = 0.3175
    imax = n//2
    for i in range(1,imax):
        irev = n - i - 1
        f =  (i+1-gam)/(n-2.0*gam+1.0)
        a[i] =  gen116( f ) 
        a[irev] =  gen116( 1.0-f )
    return a

def NormData( VerticalData ):
    yi = numpy.argsort( VerticalData )
    n = VerticalData.size
    h = gen117(n)
    x = numpy.zeros( n )
    for i in range( n ):
        x[yi[i]] = h[i]
    return x

invert = numpy.linalg.inv
def matvec(m,v):
    return numpy.inner(m,v)

def celldim(cel):
    a,b,c = splitmat(cel)
    len_a=absvec(a)
    a=a/len_a
    len_b=absvec(b)
    b=b/len_b
    len_c=absvec(c)
    c=c/len_c
    if len_a and len_b and len_c:
        alpha=numpy.arccos(numpy.add.reduce(b*c))
        beta= numpy.arccos(numpy.add.reduce(c*a))
        gamma=numpy.arccos(numpy.add.reduce(a*b))
        return (len_a,len_b,len_c,alpha,beta,gamma)
    else:
        raise error("Zero axis length")

def splitmat(m):
    """Split a matrix into its 3 columns"""
    return m[:,0],m[:,1],m[:,2]

cos=numpy.cos
sin=numpy.sin
def reccel(cel):
    """Calculate reciprocal cell parameters from direct cell parameters in a 6-tuple"""
    a,b,c,alpha,beta,gamma=cel[0],cel[1],cel[2],cel[3],cel[4],cel[5]
    alphas=numpy.arccos((cos(beta)*cos(gamma)-cos(alpha))/(sin(beta)*sin(gamma)))
    betas=numpy.arccos((cos(alpha)*cos(gamma)-cos(beta))/(sin(alpha)*sin(gamma)))
    gammas=numpy.arccos((cos(alpha)*cos(beta)-cos(gamma))/(sin(alpha)*sin(beta)))
    astar=1./(a*sin(beta)*sin(gammas))
    bstar=1./(b*sin(alpha)*sin(gammas))
    cstar=1./(c*sin(alpha)*sin(betas))
    return astar,bstar,cstar,alphas,betas,gammas

def daxdir(cel):
    """Make standard-orientation direct cell matrix from cell parameters
    """
    dm=numpy.zeros((3,3),'d')

    dm[0,0] = cel[0]
    dm[0,1] = 0.0
    dm[0,2] = 0.0

    dm[1,0] = cel[1]*cos(cel[5])
    dm[1,1] = cel[1]*sin(cel[5])
    dm[1,2] = 0.0

    dm[2,0] = cel[2]*cos(cel[4])
    dm[2,1] = (cel[1]*cel[2]*cos(cel[3])-dm[1,0]*dm[2,0]) /dm[1,1]
    dm[2,2] = cel[2]**2 - dm[2,0]**2 - dm[2,1]**2

    if dm[2,2] < 0.01:
        printcell('error:',cel)
        raise UserWarning('daxdir error')
    else:
        dm[2,2] = numpy.sqrt(dm[2,2])

#    return numpy.transpose(dm)
    return dm

def printcell(s,cell):
    newline('%s: %.5f %.5f %.5f %.2f %.2f %.2f'%(s,cell[0],cell[1],cell[2],cell[3]/DEGREES,cell[4]/DEGREES,cell[5]/DEGREES))

def bracefloat(s):
    return float(s.partition('(')[0])

def operator(s):
    newline(s)
    line = raw_input('type continue ')
    if line == 'q':
        raise

def alias(s):
# for more information: search for mathtext. Note \/ is thin space
    aliases = { 
        'iobs'            : r'$I_{obs}$',
        'logiobs'         : r'$\log(I_{obs})$',
        'sqrtiobs'        : r'$\sqrt{(I_{obs})}$',
        'icalc'           : r'$I_{calc}$',
        'sigma'           : r'$\sigma$',
        'logsigma'        : r'$\log(\sigma)$',
        'iobs/sigma'      : r'$I_{obs}/\sigma$',
        'ioversig'        : r'$I_{obs}/\sigma$',
        'sinthl'          : r'$\sin(\theta)/\lambda$',
        'theta'           : r'$\theta$',
        'reso'            : r'$resolution$',
        'd'               : r'$resolution$',
        'delta/sigma'     : r'$\Delta/\sigma$',
        'delta/sigma vs logI'   : r'$\Delta/\sigma\/vs\/\log(I_{obs})$',
        'delta/sigma vs sinthl' : r'$\Delta/\sigma\/vs\/\sin(\theta)/\lambda$',
        'delta/sigma vs d'      : r'$\Delta/\sigma\/vs\/resolution$',
        'delta/sigma vs theta'  : r'$\Delta/\sigma\/vs\/\theta$',
        'bariobs' : r'$I_{a}/I_{b}\/vs\/I_{a}$',
        'barsinthl' : r'$I_{a}/I_{b}\/vs\/\sin(\theta)/\lambda$',
        'ccsinthl' : r'$cc\/vs\/\sin(\theta)/\lambda$',
        'ccreso' : r'$cc\/vs\/resolution$',
    }
    if s in aliases:
        return aliases[s]
    else:
        return s

def strcount(n,s):
    if n==0:
        t='no '+s.replace('^','s')
    elif n==1:
        t='one '+s.replace('^','')
    else:
        t='%d %s'%(n,s.replace('^','s'))
    return t

def hklindices(refl):
    if hasattr(refl,'m') :
        hkl = (refl.h, refl.k, refl.l, refl.m, refl.n, refl.o)
    else:
        hkl = (refl.h, refl.k, refl.l)
    return hkl
    
class PLOTrecord(object):
    def __init__(self,rec=None):
        self.rec=rec

    def Defaults(self):
        self.rec = projtls.Record()
        self.type = None

    def scatter(self, hor, ver, hklp=None, name=None, samescale=True, label=None):
        self.Defaults()
        self.type = 'scatter'
        self.hor = hor
        self.ver = ver
        if samescale:
            self.hmin = min(numpy.amin(hor),numpy.amin(ver))
            self.hmax = max(numpy.amax(hor),numpy.amax(ver))
            self.vmin = self.hmin
            self.vmax = self.hmax
#            self.minmax = numpy.array( [self.hmin,self.vmax] )
        else:
            self.hmin = numpy.amin(hor)
            self.hmax = numpy.amax(hor)
            self.vmin = numpy.amin(ver)
            self.vmax = numpy.amax(ver)
#            self.minmax = numpy.array( [self.hmin,self.vmax] )

        self.hminmax = numpy.array( [self.hmin,self.hmax] )
        self.vminmax = numpy.array( [self.vmin,self.vmax] )
        if name.startswith('delta/sigma'): 
            self.vminmax = numpy.array( [0.0,0.0] )
        if hklp:
            self.hklp = hklp
        else:
            self.hklp = [ (i,0,0) for i in range(len(hor))]
        self.name = name
        self.label = label

    def scatter2(self, hor1, ver1, label1, hor2, ver2, label2, hklp=None, name=None, samescale=True):
        self.Defaults()
        self.type = 'scatter2'
        self.hor1 = hor1
        self.ver1 = ver1
        self.label1 = label1
        self.hor2 = hor2
        self.ver2 = ver2
        self.label2 = label2
        if samescale:
            self.hmin = min(numpy.amin(hor1),numpy.amin(ver1),numpy.amin(hor2),numpy.amin(ver2))
            self.hmax = max(numpy.amax(hor1),numpy.amax(ver1),numpy.amax(hor2),numpy.amax(ver2))
            self.vmin = self.hmin
            self.vmax = self.hmax
#            self.minmax = numpy.array( [self.hmin,self.vmax] )
        else:
            self.hmin = numpy.amin(hor1)
            self.hmax = numpy.amax(hor1)
            self.vmin = numpy.amin(ver1)
            self.vmax = numpy.amax(ver1)
#            self.minmax = numpy.array( [self.hmin,self.vmax] )

        self.hminmax = numpy.array( [self.hmin,self.hmax] )
        self.vminmax = numpy.array( [self.vmin,self.vmax] )
        if name.startswith('delta/sigma'): 
            self.vminmax = numpy.array( [0.0,0.0] )
            
        if hklp:
            self.hklp = hklp
        else:
            self.hklp = [ (i,0,0) for i in range(len(hor1))]
        self.name = name

    def bar(self, bins, averages, name=None):
        self.Defaults()
        self.type = 'bar'
        self.hor = bins
        self.ver = averages
        self.hmin  = 0.0
        self.hmax = bins.size
        self.vmin = min(0.0,numpy.amin(averages))
        self.vmax = numpy.amax(averages)
        self.hminmax = numpy.array( [self.hmin,self.hmax] )
        self.vminmax = numpy.array( [self.vmin,self.vmax] )
        self.name = name

class Reflrec(object):
    '''reflection record'''

    def __init__(self,items,rmat=None,wavelength=None):
        for value,name,func in items:
            if len(value) and not value.isspace():
                setattr(self,name,func(value))
        if hasattr(self,'h') and hasattr(self,'k') and hasattr(self,'l'):
            self.hkl='%d %d %d'%(self.h,self.k,self.l)
        if hasattr(self,'icalc') and hasattr(self,'iobs'):
            self.delta = self.iobs - self.icalc
        if hasattr(self,'sigma') and hasattr(self,'iobs') and hasattr(self,'icalc'):
            p = (max(self.iobs,0.0) + 2*self.icalc)/3.0
            self.wsig = numpy.sqrt( self.sigma**2 + (weighta*p)**2 + weightb*p)
        if CHANGESIGMA:
            if hasattr(self,'refsig'):
                self.sigma = self.refsig
            elif hasattr(self,'wsig'):
                self.sigma = self.wsig
        if hasattr(self,'delta') and hasattr(self,'sigma'):
            self.delsig = self.delta/self.sigma
            self.delsig2 = self.delta/(self.sigma**2)
        if rmat is not None:
            if self.h != 0 or self.k != 0 or self.l != 0:
                v=numpy.array([self.h,self.k,self.l])
                self.sinthl=float(0.5*absvec( matvec(rmat,v) ))
                self.d=0.5/self.sinthl
                if wavelength:
                    self.theta=numpy.arcsin(wavelength*self.sinthl)


    def value(self,item):
        a=getattr(self,item)
        if isinstance(a,float):
            if item in ['theta']:
                a = a/DEGREES
            s = '%.2f'%a
        else:
            s = str(a)
        return s
    
    def legenda(self):
        # return a list of tuples (name,value) in a preferred order
        preferred = [ 'hkl', 'iobs', 'sigma' ]
        ignore = []
        if 'hkl' in self.__dict__ :
            ignore=['h','k','l']
        items = []
        for item in preferred:
            if hasattr(self,item):
                items.append(item)
        for item in sorted(self.__dict__):
            if item not in items and item not in ignore:
                items.append(item)
        s = [ (item,self.value(item)) for item in items ]
        return s

    def __str__(self):
        s = []
        for item in self.legenda():
            s.append('%s=%s'%(item))
        return ' '.join(s)

    def __cmp__(self,other):
        return cmp(self.iobs,other.iobs)

class UniqueReflection(object):
    """ Unique reflection record
    """
    def __init__(self,refl):
        self.hkl = hklindices(refl)
        for f in ['sinthl','d','theta']:
            if hasattr(refl,f):
                setattr(self,f,getattr(refl,f))
        self.iobs = []
        self.sigma = []
        self.ioversig = []
        if hasattr(refl,'icalc'):
            self.icalc = []

#    def add(self, iobs, sigma, icalc=None):
#        self.iobs.append(iobs)
#        self.sigma.append(sigma)
#        self.ioversig.append(iobs/sigma)
#        if icalc is not None:
#            self.icalc.append(icalc)

    def add(self, refl):
        self.iobs.append(refl.iobs)
        self.sigma.append(refl.sigma)
        self.ioversig.append(refl.iobs/refl.sigma)
        if hasattr(refl,'icalc'):
             self.icalc.append(refl.icalc)

    def ave(self,what):
        a = getattr(self,what)
        if len(a) > 0:
            return numpy.mean(a)
        else:
            return None

    def normalize(self):
        self.iobs = [ self.ave('iobs') ]
        self.sigma = [ self.ave('sigma') ]
        self.ioversig = [ self.ave('ioversig') ]
        if hasattr(self,'icalc'):
            self.icalc = [ self.ave('icalc') ]

    def shortlegenda(self):
        return self.legenda(ignore=['iobs','sigma','ioversig','icalc'])

    def value(self,item):
        a=getattr(self,item)
        if isinstance(a,float):
            if item in ['theta']:
                a = a/DEGREES
            s = '%.2f'%a
        elif isinstance(a,list):
            r = []
            for b in a:
                if isinstance(b,float):
                    if item in ['theta']:
                        a = a/DEGREES
                    r.append('%.2f'%b)
                else:
                    r.append(str(b))
                s = ','.join(r)
        else:
            s = str(a)
        return s
        
    def legenda(self,ignore=None):
        # return a list of tuples (name,value) in a preferred order
        preferred=['hkl','d','sinthl','theta','iobs','sigma','ioversig','icalc']
        if ignore is None:
            ignore = []
        items = []
        for item in preferred:
            if hasattr(self,item) and item not in ignore:
                items.append(item)
        for item in sorted(self.__dict__):
            if item not in items and item not in ignore:
                newline(item,item in ignore)
                items.append(item)
        s = [ (item,self.value(item)) for item in items ]
        return s

    def __str__(self):
        s = []
        for item in self.legenda():
            s.append('%s=%s'%(item))
        return ' '.join(s)

def normprob(IO,IC,S):
    delta = IO-IC
    sigma = S
    delsig = delta/sigma

    res = scipy.stats.probplot( delsig, fit=True, dist='norm')  # compute
    osm,osr = res[0]
    slope,intercept,r = res[1]
    newline('Normal probability (%s): slope=%.4f intercept=%.4f r=%.4f'%(plural.reflections(IO),slope,intercept,r))
    yi = numpy.argsort( delsig )
    nrefl = len(IO)
    x = numpy.zeros( nrefl )
    for i in range( nrefl ):
        x[yi[i]] = osm[i]
    y = delsig
    label = 'Slope=%.4f Intercept=%.4f'%(slope,intercept)
    return x,y,label

def ReadFCF(filename,rmat,wavelength=None):
    # create empty reflection list
    refls=[]
    # create a list of requested items
    requests = []
    f = open(filename,'r')
    WithinLoop = False
    WithinReflLoop = False
    cell = []
    newline()
    newline('reading',filename)
    for line in f:
        line = line.strip()
        n = line.find('#')
        if n >= 0:
            line = line[:n]
        if len(line) == 0:
            WithinLoop = False
            WithinReflLoop = False
        elif line[:5] == 'loop_':
            WithinLoop = True
        elif line[:6] == '_cell_':
            if rmat is None:
                items = line.split()
                if items[0][6:12] == 'length':
                    cell.append(bracefloat(items[1]))
                elif items[0][6:11] == 'angle':
                    cell.append(bracefloat(items[1])*DEGREES)
                if len(cell) == 6:
# N.B. this collects the first set of cell parameters
                    dmat = daxdir(cell)
                    rmat = invert(dmat)
                    # newline(rmat)
        elif line[:7] == '_refln_' :
            WithinReflLoop = WithinLoop
            if line[7:] == 'index_h':
                requests.append(('h',int))
            elif line[7:] == 'index_k':
                requests.append(('k',int))
            elif line[7:] == 'index_l':
                requests.append(('l',int))
            elif line[7:] == 'F_squared_calc':
                requests.append(('icalc',float))
            elif line[7:] == 'F_squared_meas':
                requests.append(('iobs',float))
            elif line[7:] == 'F_squared_sigma':
                requests.append(('sigma',float))
            elif line[7:] == 'observed_status':
                requests.append(('status',str))
            elif line[7:] == 'XD_refine_code':
                requests.append(('status',int))
            else:
                newline('  ignoring %s'%line)
                requests.append((None,None))
        elif line[:7] ==  '_shelx_' :
            WithinReflLoop = WithinLoop
            if line[7:] == 'refinement_sigma':
                requests.append(('refsig',float))
            else:
                newline('  ignoring %s'%line)
#                requests.append((None,None))
        elif WithinReflLoop:
            items = line.split()
            fields = []
            for item,request in zip(items,requests):
                if request[0] != None:
                    fields.append( ( item, request[0], request[1]) )
            new = Reflrec(fields,rmat=rmat,wavelength=wavelength)
            refls.append(new)
        
    f.close
    newline('  %s from %s'%(plural.reflections(refls),filename))
    root,ft=os.path.splitext(filename)
    if ft == '.fco' and not args.ignorestatus:
        new = []
        nstatus = 0
        for refl in refls:
            if refl.status not in [0,'o']:   # zero and letter o
                nstatus +=1
            else:
                new.append(refl)
        if nstatus > 0:
            newline('  %s rejected on status'%(plural.reflections(nstatus)))
        refls = new[:]
    if len(cell) == 6:
        dmat = daxdir(cell)
        printcell('  cell from %s'%filename,cell)
    return rmat,refls

def ReadHKLF4(filename,rmat,wavelength=None):
    # create empty reflection list
    refls=[]
    # create a list of requested items
    requests = [ ('h',int), ('k',int), ('l',int), ('iobs',float), ('sigma',float) ]
    f = open(filename,'r')
    newline('reading %s'%filename)
    for line in f:
        if len(line) <= 1:
            break
        items = [ line[0:4], line[4:8], line[8:12], line[12:20], line[20:28] ]
        fields = []
        for item,request in zip(items,requests):
            fields.append( (item,request[0],request[1]) )
        new = Reflrec(fields,rmat=rmat,wavelength=wavelength)
        if new.h == 0 and new.k == 0 and new.l == 0:
            break
        refls.append(new)
    f.close
    newline('  %s'%plural.reflections(refls))
    return rmat,refls

def ReadHKLIN(filename,rmat,wavelength=None):
    # create empty reflection list
    refls=[]
    # create a list of requested items
    requests = [ ('h',int), ('k',int), ('l',int), ('iobs',float), ('sigma',float) ]
    f = open(filename,'r')
    p0 = re.compile(r'\s*([-0-9]+)\s*([-0-9]+)\s*([-0-9]+)\s*([-0-9]+\.[0-9]+)\s*([-0-9]+\.[0-9]+)')
    newline('reading %s'%filename)
    for line in f:
        if len(line) <= 1:
            break
        m0 = p0.match(line)
#        newline(line)
        if m0:
            items = m0.groups()
#        items = [ line[0:5], line[5:10], line[10:15], line[15:25], line[25:35] ]
        fields = []
        for item,request in zip(items,requests):
            fields.append( (item,request[0],request[1]) )
        new = Reflrec(fields,rmat=rmat,wavelength=wavelength)
        if new.h == 0 and new.k == 0 and new.l == 0:
            break
        refls.append(new)
    f.close
    newline('  %s'%strcount(len(refls),'reflection^'))
    return rmat,refls

def ReadHK6(filename,rmat,wavelength=None):
    # create empty reflection list
    refls=[]
    # create a list of requested items
    requests = [ ('h',int), ('k',int), ('l',int), ('m',int), ('n',int), ('o',int), ('iobs',float), ('sigma',float) ]
    f = open(filename,'r')
    newline('reading %s'%filename)
    for line in f:
        items = [ line[0:4], line[4:8], line[8:12], line[12:16], line[16:20], line [20:24], line[24:32], line[32:40] ]
        fields = []
        for item,request in zip(items,requests):
            fields.append( (item,request[0],request[1]) )
        new = Reflrec(fields,rmat=rmat,wavelength=wavelength)
        refls.append(new)
    f.close
    newline('  %s'%strcount(len(refls),'reflection^'))
    return rmat,refls

def ReadRefls(filename,rmat,wavelength=None):
    funcs={
        '.hkl':ReadHKLF4,
        '.hklin':ReadHKLIN,
        '.hklv':ReadHKLF4,
        '.sad':ReadHKLF4,
        '.raw':ReadHKLF4,
        '.hkp':ReadHKLF4,
        '.hk6':ReadHK6,
        '.ram':ReadHK6,
        '.fcf':ReadFCF,
        '.fco':ReadFCF
        }
    root,ft=os.path.splitext(filename)
    try:
        rmat,refls = funcs[ft](filename,rmat,wavelength=wavelength)
    except KeyError:
        newline('unknown filetype %s in %s'%(ft,filename))
        newline('possible types are: %s'%(' '.join(funcs)))
        raise IOError('bad filetype')

    if len(refls) > 0:
        if CHANGESIGMA:
            if hasattr(refls[0],'refsig'):
                newline('  using refined sigma')
            elif weighta != 0.0 or weightb != 0.0:
                newline('  using weight corrected sigma')
            else:
                newline("  original sigma")
        else:
            newline("  original sigma")
    return rmat,refls

def Unique(refls,domerge=True,allowNegative=False):
    """
    sum all unique reflections (if domerge)
    sigma should be positive (because of I/sigma)
    iobs should be positive (because of log(i), later)
    """
    unique = {}
    nbadi = 0
    nbads = 0

    for refl in refls:
        if refl.sigma > 0:
            if allowNegative or refl.iobs > 0.0:
                hkl = hklindices(refl)
                if hkl not in unique:
                    unique[hkl] = UniqueReflection(refl)
                unique[hkl].add(refl)
            else:
                nbadi +=1
        else:
            nbads +=1
    if nbadi > 0:
        newline('  %s ignored because iobs is not positive'%(plural.reflections(nbadi)))
    if nbads > 0:
        newline('  %s ignored because sigma is not positive'%(plural.reflections(nbads)))
    newline('  %s'%(plural.nplural(unique,'unique reflection')))
    # normalize

    message = "merging intensities and sigma's for reflections with equal indices"
    if domerge:
        newline('  '+message)
        for hkl in unique:
            unique[hkl].normalize()
    else:
        newline('  not '+message)
        ni = sum( len(unique[hkl].iobs) for hkl in unique)
        newline('  %d intensities'%ni)

    if args.verbose > 1:
        np = 0
        for hkl in unique:
            newline(hkl,unique[hkl])
            np += 1
            if np > DEBUGMAXREFLS:
                newline('reflection list stopped after %s reflections'%np)
                break
    return unique

def main():
    import numpy

    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib import rc
    import matplotlib.widgets as wid
    global mode
    global ndxp
    global ndxh
    global weighta
    global weightb
    global onscreen

    info = [ 'iobs', 'logiobs', 'sqrtiobs', 'sigma', 'logsigma', 'iobs/sigma' ] # optional icalc will be added later
    
    def GetBins(n,d):
        scr = numpy.array(d)
        scrs = numpy.sort(scr)
        if args.verbose > 1:
            newline('first, last=',scrs[0],scrs[-1])
        # make sure the upper limit will be larger than the largest value
        scrs[-1] = scrs[-1] + EXTRAUPPER
        # note, bins are interpreted as upper limits.
        # generate interval limits, one more than the number of bins
        x = numpy.linspace(0, scr.size-1, n+1)
        # the values at positions x contain the bin limits
        # the first point in x will be ignored (it's a lower limit)
        bins = numpy.zeros(n)
        for g in range(n):
            bins[g] = scrs[ int(x[g+1]) ]
        if args.verbose > 1:
            newline('bins=',bins)
        inds = numpy.digitize(scr,bins)
        return bins,inds

    def margins( hmin,hmax,vmin,vmax):
#        newline('margins1: %f %f %f %f'%(hmin,hmax,vmin,vmax))
        rh = hmax-hmin
        hmin = hmin-margin*rh
        hmax = hmax+margin*rh
        rv = vmax-vmin
        vmin = vmin-margin*rv
        vmax = vmax+margin*rv
        ax.set_xlim(hmin,hmax)
        ax.set_ylim(vmin,vmax)
#        newline('margins2: %f %f %f %f'%(hmin,hmax,vmin,vmax))

    def labelit(a):
        size='large'
        # draw title and label the axes
        a.set_title(almodes[mode],size=size)
        a.set_xlabel(filelist[0],size=size)
        if modes[mode] == 'bariobs':
            a.set_xlabel( alias('iobs')+' '+filelist[0],size=size)
#            a.set_ylabel( '$I_{'+filelist[1]+'} / I_{'+filelist[0]+'}',size=size)
            a.set_ylabel( r'$I_{%s}\//\/I_{%s}$'%(filelist[0],filelist[1]),size=size)
        elif modes[mode] == 'barsinthl':
            a.set_xlabel( alias('sinthl'),size=size)
#            a.set_ylabel( filelist[1]+'/'+filelist[0],size=size)
            a.set_ylabel( r'$I_{%s}\//\/I_{%s}$'%(filelist[0],filelist[1]),size=size)
        elif modes[mode] == 'ccsinthl':
            a.set_xlabel( alias('sinthl'),size=size)
            a.set_ylabel( 'cc',size=size)
        elif modes[mode] == 'ccreso':
            a.set_xlabel( alias('reso'),size=size)
            a.set_ylabel( 'cc',size=size)
        elif modes[mode] == 'normprobold':
            a.set_xlabel( alias('old prob'),size=size)
            a.set_ylabel( alias('delta/sigma'),size=size)
        elif modes[mode] == 'normprob':
            a.set_xlabel( alias('prob'),size=size)
            a.set_ylabel( alias('delta/sigma'),size=size)
        elif modes[mode] == 'normprob2':
            a.set_xlabel( alias('prob2'),size=size)
            label = '(%s-%s)/%s'%(alias('iobs'),alias('icalc'),alias('sigma'))
            a.set_ylabel(label,size=size)
        elif modes[mode].startswith('delta/sigma'):
            a.set_ylabel( alias('delta/sigma'),size=size)
            if modes[mode].endswith(' logI'):
                a.set_xlabel( alias('logiobs')+' '+filelist[0],size=size)
            elif modes[mode].endswith(' sinthl'):
                a.set_xlabel( alias('sinthl'),size=size)
            elif modes[mode].endswith(' d'):
                a.set_xlabel( alias('reso'),size=size)
            elif modes[mode].endswith(' theta'):
                a.set_xlabel( alias('theta'),size=size)
        elif modes[mode] != 'iobs/sigma' :
            a.set_ylabel('%.2f*%s'%(scale,filelist[1]),size=size)
        else:
            a.set_ylabel(filelist[1],size=size)

    def CreateTable(bottom,labels,title=None):
        aaa = []
        if title is not None:
            aaa.append( (None, title) )
        aaa.extend(labels)
        ll = len(aaa)
        height = ll*args.charheight
        bottom = bottom - height - GAP
        newax = plt.axes([0.01, bottom, SUBWINDOWWIDTH , height])
        newax.set_navigate(False)
        newax.set_xticks([])
        newax.set_yticks([])
        dy = 1./(ll)
        ys = numpy.linspace(1-0.5*dy, 0.5*dy, ll)
        xx = [ GAP, 1.0-GAP ] 
        result = []
        for y,label in zip(ys,aaa):
            if label[0] is None:
                parts = label[1].partition(':')
                if parts[1] == '':
                    column1 = ( newax.text(0.5, y, parts[0],
                                           horizontalalignment='center',
                                           verticalalignment='center',
                                           color='blue'))
                    column2 = ( newax.text(xx[1], y, '',
                                           horizontalalignment='right',
                                           verticalalignment='center', 
                                           color='blue'))
                else:
                    column1 = ( newax.text(xx[0], y, parts[0],
                                           horizontalalignment='left',
                                           verticalalignment='center',
                                           color='blue'))
                    column2 = ( newax.text(xx[1], y, parts[2], 
                                           horizontalalignment='right',
                                           verticalalignment='center',
                                           color='blue'))
            else:
                s = alias(label[0])
                column1 = ( newax.text(xx[0], y, s, verticalalignment='center' ))
                column2 = ( newax.text(xx[1], y, label[1], 
                                       horizontalalignment='right',
                                       verticalalignment='center') )
            result.append( (label[0],column2) )
        return bottom,result

    def on_pick(event):
        """
        If one dot is picked, highlight the reflection, 
        and display reflection information in the table.
        ndxp = index of point
        ndxh = hkl name of ndxp
        ndxp and ndxh are global (if modes are switched, the ndxh reflection and ndxp points have 
        to be highlighted also).
        If more than one dot is within the pickable area, no 
        highlighting occurs, but the reflections will be printed.
        """
        global ndxp
        global ndxh
        if args.verbose > 1:
            newline('on pick:',event.ind)
        newline()
        lastndxp = ndxp
        lastndxh = ndxh
        if plotdata[mode].type == 'scatter':
            if len(event.ind) > 0:
                h = plotdata[mode].hor
                v = plotdata[mode].ver
                refh = event.mouseevent.xdata
                refv = event.mouseevent.ydata
                d = [ (h[i]-refh)**2 + (v[i]-refv)**2 for i in event.ind ]
                ndxp = event.ind[ d.index( min(d) ) ]
                ndxh = plotdata[mode].hklp[ ndxp ]
                if args.verbose > 1:
                    newline('d=',d)
                    newline('min=',min(d))
                    newline('ndxp=',ndxp)
                    newline('ndxh=',ndxh)
#                L2.set_visible(True)
                        
                if ndxp == lastndxp:
                    ndxp = -1
                    ndxh = (-1,0,0)
                drawhkl()
                printhkl()
        elif plotdata[mode].type == 'scatter2':
            if len(event.ind) > 0:
                h = plotdata[mode].hor1
                v = plotdata[mode].ver1
                refh = event.mouseevent.xdata
                refv = event.mouseevent.ydata
                d = [ (h[i]-refh)**2 + (v[i]-refv)**2 for i in event.ind ]
                ndxp = event.ind[ d.index( min(d) ) ]
                ndxh = plotdata[mode].hklp[ ndxp ]
                if args.verbose > 1:
                    newline('d=',d)
                    newline('min=',min(d))
                    newline('ndxp=',ndxp)
                    newline('ndxh=',ndxh)
#                L2.set_visible(True)
                        
                if ndxp == lastndxp:
                    ndxp = -1
                    ndxh = (-1,0,0)
                drawhkl()
                printhkl()
        elif plotdata[mode].type == 'bar':
            newline('cannot pick bar')

    def switchfunc(label):
        """
        switch to another plot
        """
        global mode
        global onscreen
        ax.cla()
        if args.verbose > 1:
            newline('switchfunc1:',label,mode)
        onscreen = {}
        mode = almodes.index(label)
        if args.verbose > 1:
            newline('switchfunc2:',mode)
        pd = plotdata[mode]
        labelit(ax)
        if plotdata[mode].type == 'scatter':
            # a red line from min,min to max,max
            L0, = ax.plot(pd.hminmax, pd.vminmax, 'r-')
#            L0, = ax.plot(pd.minmax, 'r-')
            # the dots at (x,y). pickable
            L1, = ax.plot(pd.hor, pd.ver, 'b.', markersize=dotsize, picker=True,pickradius=5)
            # an invisible highlighted dot.
            ndxp = 0
            ndxh = (0,0,0)
            if args.merge != 12:
                L2, = ax.plot(pd.hor[0], pd.ver[0],
                              marker='o',
                              linestyle='None',
                              markersize=2*dotsize,
                              markerfacecolor='yellow',
                              markeredgecolor='black',
                              visible=False)
                onscreen['multi']=L2
            L3, = ax.plot(pd.hor[0], pd.ver[0],
                          'r.',
                          markersize=2*dotsize,
                          visible=False)
            onscreen['hkl']=L3

            ndxp = -1 # no highlight for the moment
            ndxh = (-1,0,0) # no highlight for the moment
            margins( pd.hmin, pd.hmax, pd.vmin, pd.vmax )
            if pd.label:
                ax.text(0.99,0.01,
                        pd.label,
                        color='r',
                        horizontalalignment='right',
                        verticalalignment='bottom',
                        transform=ax.transAxes)

            drawhkl()
        if plotdata[mode].type == 'scatter2':
            # a red line from min,min to max,max
            L0, = ax.plot(pd.hminmax, pd.vminmax, 'r-')
#            L0, = ax.plot(pd.minmax, 'r-')
            # the dots at (x,y). pickable
            L1, = ax.plot(pd.hor1, pd.ver1, 'b.', markersize=dotsize, picker=True,pickradius=5)
            # an invisible highlighted dot.
            ndxp = 0
            ndxh = (0,0,0)
            if args.merge != 12:
                L2, = ax.plot(pd.hor1[0], pd.ver1[0],
                              marker='o',
                              linestyle='None',
                              markersize=2*dotsize,
                              markerfacecolor='yellow',
                              markeredgecolor='black',
                              visible=False)
                onscreen['multi']=L2
            L3, = ax.plot(pd.hor1[0], pd.ver1[0],
                          'r.',
                          markersize=2*dotsize,
                          visible=False)
            onscreen['hkl']=L3

            L4, = ax.plot(pd.hor2, pd.ver2, 'g.', markersize=dotsize, picker=True,pickradius=5)
            # an invisible highlighted dot.

            ndxp = -1 # no highlight for the moment
            ndxh = (-1,0,0) # no highlight for the moment
            margins( pd.hmin, pd.hmax, pd.vmin, pd.vmax )
            ax.text(0.99,0.01+args.charheight,
                    pd.label1,
                    color='b',
                    horizontalalignment='right',
                    verticalalignment='bottom',
                    transform=ax.transAxes)
            ax.text(0.99,0.01,
                    pd.label2,
                    color='g',
                    horizontalalignment='right',
                    verticalalignment='bottom',
                    transform=ax.transAxes)

            drawhkl()
        elif plotdata[mode].type == 'bar':
            B0=ax.bar(numpy.arange(pd.ver.size),pd.ver)
            if args.verbose > 0 or True:
                #for h,v in zip(pd.hor,pd.ver):
                #    newline('%.4f %.4f'%(h,v))
#            ax.set_xlabel( 'hor', size='xx-large' )
#            ax.set_ylabel( 'ver', size='xx-large' )
                hormax = len(str(int(max(pd.hor))))
                vermax = len(str(int(max(pd.ver))))
                #countmax = len(str(max(pd.count)))
                #newline(hormax,vermax,countmax)
                hormant = 2
                vermant = 2
                form =  '%%%ds %%%ds'%(hormax+1,hormant+vermax+1)
                newline(form%('x','y'))
                form = '%%%d.%df %%%d.%df'%(hormax+1+hormant,hormant,vermax+1+vermant,vermant)
                #newline(form)
                for h,v in zip(pd.hor,pd.ver):
                    newline(form%(h,v))
                label = 'mean %.4f variance %.4f'%(numpy.mean(pd.ver),numpy.var(pd.ver))
                newline(label)
                ax.text(0.0,1.00+0.3*args.charheight,
                         label,
                         color='r',
                         horizontalalignment='left',
                         verticalalignment='bottom',
                         transform=ax.transAxes)

            margins( pd.hmin, pd.hmax, pd.vmin, pd.vmax )
            name = plotdata[mode].name
            if name.endswith( ('sinthl','reso') ):
                labels = []
                h = pd.hor.copy()
                h[-1] = h[-1] - EXTRAUPPER
                for f in h:
                    labels.append('%.2f'%f)
                ax.set_xticks(numpy.arange(pd.ver.size)+0.9)
                ax.set_xticklabels(labels)
            #newline(mode)
        plt.draw()

    def createButtons(bottom):
        buttons = {}
        height = 0.04
        width = 0.06
        x = 0.01
        buttonax = plt.axes([x, bottom, width, height])
        button = wid.Button(buttonax, 'Quit')
        button.on_clicked(quitbuttonfunc)
        buttons['quit'] = button
        x = x + width
        buttonax = plt.axes([x, bottom, width, height])
        button = wid.Button(buttonax, 'All-%s'%args.savetype)
        button.on_clicked(savebuttonfunc)
        buttons['save'] = button
        x = x + width
        buttonax = plt.axes([x, bottom, width, height])
        button = wid.Button(buttonax, 'HKL')
        button.on_clicked(hklbuttonfunc)
        buttons['hkl'] = button
        bottom = bottom + height + GAP
        return bottom,buttons

    def createRadios(bottom):
        def change_buttons(r, radiusfactor,circlex=0.05,labelx=0.1):
            "Resize and move all radio buttons in `r` collection by fractions `f`"
            for c in r.circles:
                c.set_radius(c.get_radius()*radiusfactor)
                c.center = (circlex,c.center[1])
            for c in r.labels:
#                c.set_fontsize(10)
                old = c.get_position()
                c.set_position((labelx,old[1]))
            rpos = r.ax.get_position().get_points()
            fh = fig.get_figheight()
            fw = fig.get_figwidth()
            rscale = (rpos[:,1].ptp() / rpos[:,0].ptp()) * (fh / fw)
            for c in r.circles:
                c.height /= rscale

        radios = {}
        height = 1.1*len(almodes)*args.charheight
        radioax = plt.axes([0.01, bottom, SUBWINDOWWIDTH, height])
        radio = wid.RadioButtons(radioax, almodes, active=mode)
        change_buttons(radio,0.6) # shrink your buttons in half
        radio.on_clicked(switchfunc)
        radios['switch'] = radio
        bottom = bottom + height + GAP
        return bottom,radios

    def printhkl():
        if ndxh[0] >= 0:
            ref0 = uniqueA[hkls[ndxh[0]]]
            ref1 = uniqueB[hkls[ndxh[0]]]
            newline('%s: %s' % ( filelist[0], ref0 ))
            newline('%s: %s' % ( filelist[1], ref1 ))

            if modes[mode] in info:
                # plotvalue returns useful data only for modes in info
                len0 = len(ref0.iobs)
                lens = []
                for z0 in range(len0):
                    ploth = plotvalue(ref0,modes[mode],z0,1.0)
                    if ploth is not None:
                        lens.append(len('%.2f'%ploth))
                    else:
                        lens.append(0)
                size0 = max(lens)

                len1 = len(ref1.iobs)
                lens = []
                for z1 in range(len1):
                    plotv = plotvalue(ref1,modes[mode],z1,1.0)
                    if plotv is not None:
                        lens.append(len('%.2f'%plotv))
                    else:
                        lens.append(0)
                size1 = max(lens)
                newline('{:^{width}}'.format(modes[mode],width=size0+size1+1))
                newline('{:>{width0}} {:>{width1}}'.format('hor','ver',width0=size0, width1=size1))
                for z0 in range(len0):
                    for z1 in range(len1):
                        ploth = plotvalue(ref0,modes[mode],z0,1.0)
                        plotv = plotvalue(ref1,modes[mode],z1,1.0)
                        bingo = ''
                        if len0 > 1 or len1 > 1:
                            if z0 == ndxh[1] and z1 == ndxh[2]:
                                bingo = ' *'
                        # newline('%.2f %.2f%s'%(ploth,plotv,bingo))
                        newline('{:{width0}.2f} {:{width1}.2f}{}'.format(ploth,plotv, bingo, width0=size0, width1=size1))
                        
    def plotvalue(u,m,offset,sf):
        if m == 'iobs':
            result = sf*u.iobs[offset]
        elif m == 'logiobs':
            result = numpy.log10(sf*u.iobs[offset])
        elif m == 'sqrtiobs':
            result = numpy.sqrt(sf*u.iobs[offset])
        elif m == 'sigma' :
            result = sf*u.sigma[offset]
        elif m == 'logsigma' :
            result = numpy.log10(sf*u.sigma[offset])
        elif m == 'iobs/sigma':
            result = u.ioversig[offset]
        elif m == 'icalc':
            result = sf*u.icalc[offset]
        else:
            result = None
        return result

    def drawhkl():
        """
        highlight a picked reflection, and display reflection 
        information in the table.
        """
        global onscreen
        def OneColumn(u,offset,sf,rows):
            data = []
            for m in info:
                data.append(plotvalue(u,m,offset,sf))
            for r,f in zip(rows[1:],data):
                plt.setp(r,text=r'%.2f'%f)

        def Visibility(b):
            if 'multi' in onscreen:
                L2.set_visible(b)
            if 'hkl' in onscreen:
                L3.set_visible(b)

#        newline('ndxp=',ndxp,'ndxh=',ndxh)
        if args.merge != 12 and ndxh[0] >= 0:
            if args.verbose > 0:
                newline('drawhkl: search ndxp for ndxh=',ndxh)
            pd = plotdata[mode]
            if pd.type == 'scatter':
                x = []
                y = []
                ip = []
                for i,p in enumerate(pd.hklp):
                    if p[0] == ndxh[0]:
                        x.append(pd.hor[i])
                        y.append(pd.ver[i])
                        ip.append(i)
                if len(x):
                    if args.verbose > 1:
                        newline('Change L2 and visible')
                        newline()
                    if 'multi' in onscreen:
                        L2 = onscreen['multi']
                        L2.set_xdata(x)
                        L2.set_ydata(y)
                    else:
                        if args.verbose > 1:
                            newline('Create L2')
                        L2,=ax.plot(x,y,'g.',markersize=2*dotsize)
                    onscreen['multi']=L2
#                    Visibility(True)
        if ndxp >= 0:
            #newline('drawhkl',ndxp,ndxh)
            pd = plotdata[mode]
            if pd.type == 'scatter':
                if 'hkl' in onscreen:
                    if args.verbose > 1:
                        newline('Change L3 and visible')
                    L3 = onscreen['hkl']
                    L3.set_xdata(pd.hor[ndxp])
                    L3.set_ydata(pd.ver[ndxp])
                else:
                    if args.verbose > 1:
                        newline('Create L3')
                    L3,=ax.plot(pd.hor[ndxp],pd.ver[ndxp],'r.',markersize=4*dotsize)
                    onscreen['hkl']=L3

                Visibility(True)

                hklp,ap,bp = ndxh
                for item,pos in legendaMain:
                    if item:
                        value = uniqueA[hkls[hklp]].value(item)
                        plt.setp(pos,text=value)

                for item,pos in legendaOne:
                    if item:
                        #newline('xx',item,pos)
                        value = plotvalue(uniqueA[hkls[hklp]],item,ap,1.0)
                        #newline('  ',value)
                        plt.setp(pos,text='%.2f'%value)
                for item,pos in legendaTwo:
                    if item:
                        value = plotvalue(uniqueB[hkls[hklp]],item,bp,scale)
                        plt.setp(pos,text='%.2f'%value)

            elif pd.type == 'bar':
                newline('Visibility bar')
        else:
            if 'hkl' in onscreen:
                if args.verbose > 1:
                    newline('get L3 and invisible')
                L3 = onscreen['hkl']
                if 'multi' in onscreen:
                    if args.verbose > 1:
                        newline('get L2')
                    L2 = onscreen['multi']
                Visibility(False)
        plt.draw()

    def triggerSwitch(m):
        n = almodes.index(m)
        radio = radios['switch']
        for i,p in enumerate(radio.circles):
            if i==n:
                color = radio.activecolor
            else:
                try:
                    color = radio.ax.get_facecolor()
                except AttributeError:
                    color = radio.ax.get_axis_bgcolor()
            p.set_facecolor(color)
        if radio.drawon: radio.ax.figure.canvas.draw()
        switchfunc(almodes[n])

    def quitbuttonfunc(label):
        sys.exit()

    def ableButtons(state):
        for key,value in buttons.items():
            if key != 'quit':
                value.eventson = state

    def savebuttonfunc(event):
        ableButtons(False)
        lastmode = mode
        tmpl = 'loglog-%s-%02d.'+args.savetype
        newline(tmpl)
        for button in buttons:
            buttons[button].ax.set_visible(False)

        for i,m in enumerate(almodes):
            newline(i,m)
            triggerSwitch(m)
            fname = tmpl%(filename,i+1)
            newline('save to',fname)
            plt.savefig(fname)
        for button in buttons:
            buttons[button].ax.set_visible(True)
        if i != lastmode:
            triggerSwitch(almodes[lastmode])
        ableButtons(True)
        newline('%d pictures saved'%len(modes))

    def pressrelease(event):
        if event.name == 'key_press_event':
            if event.key=='q':
                quitbuttonfunc('q-press')
            elif event.key == 'down':
                triggerSwitch(almodes[(mode+1)%len(almodes)])
            elif event.key == 'up':
                triggerSwitch(almodes[(mode-1)%len(almodes)])
            else:
                newline('ignored %s %s'%(event.name,event.key))
        else:
            newline('ignored %s %s'%(event.name,event.key))

    def getlistarray(data,what,hklset):
        #newline('lenhklset',what,len(hklset))
        l = [ getattr(data[hkl],what) for hkl in hklset] 
        if type(l[0]) == list:
            return numpy.array( [ item for sublist in l for item in sublist ] )
        else:
            return numpy.array(l)

    def getpairs(what):
        return [ (getattr(uniqueA[hkl],what),getattr(uniqueB[hkl],what)) for hkl in common] 

    def getsingles(what,func=None):
        singles = [ getattr(uniqueA[hkl],what) for hkl in common ] 
        x = []
        hklp = []
        if type(singles[0]) == list:
            for p,hor in enumerate(singles):
                for ih,h in enumerate(hor):
                    x.append(h)
                    hklp.append( (p,ih,0) )
        else:
            x = singles
            for p in range(len(singles)):
                hklp.append( (p,0,0) )

        x = numpy.array(x)
        if func:
            x = func(x)
        return x,hklp

    def getpoints(what,scale=1.0,func=None):
        pairs = getpairs(what)
        x = []
        y = []
        hklp = []
        for p,(hor,ver) in enumerate(pairs):
            for ih,h in enumerate(hor):
                for iv,v in enumerate(ver):
                    x.append(h)
                    y.append(scale*v)
                    hklp.append( (p,ih,iv) )
        x = numpy.array(x)
        y = numpy.array(y)
        if func:
            x = func(x)
            y = func(y)
        return x,y,hklp

    def hklbuttonfunc(label):
        """
        Wait for h,k,l from input. 
        Highlight available selected reflection.
        The table will be cleared if the reflection is not available.
        """
        global ndxp
        global ndxh
        name = raw_input('Enter hkl:')
        items = name.replace(',',' ').split()
        if len(items) != 3:
            newline('expected 3 numbers')
            ndxp = -1
        else:
            hkl = tuple([int(x) for x in items])
            if hkl in common:
                ndxh = (hkls.index(hkl),0,0)
                pm = plotdata[mode]
                ndxp = pm.hklp.index(ndxh)
                if args.verbose > 0:
                    newline('indexh=',ndxh)
                    newline('indexp=',ndxp)
                    newline(pm.hor[ndxp],plotdata[mode].ver[ndxp])
                drawhkl()
                printhkl()
            else:
                newline('no such reflection')
                ndxp = -1
                ndxh = (-1,0,0)
    weighta = 0.0
    weightb = 0.0
    abort = False
    
    df = ' [default=%(default)s]'

    parser = argparse.ArgumentParser(
        usage=_prog_usage,
        description=_prog_description)

    parser.set_defaults(
        verbose=0,
        dotsize=4,
        ignorestatus=False,
        mode=1,  # start with loglog
        nbin=10,
        wavelength=None,
        merge=12,
        pagehor=29.7,
        pagever=21.0,
        savetype='png',
        )

    parser.add_argument(
        '-v','--verbose',
        action='count',
        help='more output'+df)

    parser.add_argument(
        '-c','--cell',
        type=float,
        nargs=6,
        metavar='f',
        help='enter 6 cell parameters')

    parser.add_argument(
        '-d','--dotsize',
        type=int,
        metavar='n',
        help='dotsize of scatter points'+df)

    parser.add_argument(
        '--ignorestatus',
        action='store_true',
        help='ignore status flag in fco files'+df)

    parser.add_argument(
        '-m','--mode',
        type=int,
        metavar='n',
        help='start with specific plot'+df)

    parser.add_argument(
        '--merge',
        type=int,
        metavar='n',
        help='merge same reflections (0=no/1=first/2=second/12=both)'+df)

    parser.add_argument(
        '-n','--nbin',
        type=int,
        metavar='n',
        help='Number of bins in plots'+df)

    parser.add_argument(
        '--negative',
        action='store_true',
        dest='allowNegative',
        help='allow negative intensities'+df)

    parser.add_argument(
        '--pagehor',
        type=float,
        metavar='f',
        help='horizontal page size'+df)

    parser.add_argument(
        '--pagever',
        type=float,
        metavar='f',
        help='vertical page size'+df)

    parser.add_argument(
        '-r', '--rmat',
        help='read rmat')

    parser.add_argument(
        '--savetype',
        help='set filetype for save'+df)

    parser.add_argument(
        '-s','--scale',
        type=float,
        metavar='f',
        help='Use fixed scale factor'+df)

    parser.add_argument(
        '-w','--wavelength',
        type=float,
        metavar='f',
        help='radiation wavelength')

    parser.add_argument(
        'files',
        nargs=2,
        help='2 hkl files')

    global args
    args = parser.parse_args()

    dotsize = args.dotsize
    mode = args.mode
    scale = args.scale

    if args.verbose > 1:
        newline('args:',args)

    domergeA = args.merge in [1,12]
    domergeB = args.merge in [2,12]
    
    if args.rmat:
        cell = rmat.Rmat(args.rmat)
        rm = cell.GetOrientationMatrix()
        dm = invert(rm)
        printcell('cell from %s'%args.rmat,celldim(dm))
    elif args.cell:
        cell = list(args.cell)
        cell[3]=cell[3]*DEGREES
        cell[4]=cell[4]*DEGREES
        cell[5]=cell[5]*DEGREES
        dm = daxdir(cell)
        printcell('cell from commandline',cell)
        if args.verbose > 0:
            newline(dm)
        rm = invert(dm)
        if args.verbose > 1:
            newline(rm)
    else:
        dm = None
        rm =None

    filelist = []
    for f in args.files:
        if os.path.exists(f):
            filelist.append(f)
        else:
            newline('unknown option: ',f)
            abort=True

    if abort:
        raise UserWarning('unknown options')

    if len(filelist) != 2:
        raise UserWarning('need 2 filenames')

    negativeExists = False
    filename = os.path.splitext(filelist[0])[0] + '_' + os.path.splitext(filelist[1])[0]


    rm,reflsA = ReadRefls(filelist[0],rm,wavelength=args.wavelength)
    uniqueA = Unique(reflsA,domerge=domergeA,allowNegative=args.allowNegative)

    rm,reflsB = ReadRefls(filelist[1],rm,wavelength=args.wavelength)
    uniqueB = Unique(reflsB,domerge=domergeB,allowNegative=args.allowNegative)

    negativeExists = any( item.iobs <= 0.0 for item in reflsA+reflsB)
#    newline(negativeExists)
#    for refl in reflsA+reflsB:
#        if refl.iobs <= 0.0:
#            negativeExists = True
#            break

#    newline(negativeExists)
    if negativeExists and args.allowNegative:
        info = [i for i in info if not i.startswith(('log','sqrt')) ]

    if args.verbose > 1:
        newline(reflsA[0])

    modes = info[:]

    withreso = hasattr(reflsA[0],'sinthl')
    withtheta = hasattr(reflsA[0],'theta')
    if args.merge == 12:
        modes.append('delta/sigma vs logI')
        if withreso:
            modes.append('delta/sigma vs sinthl')
            modes.append('delta/sigma vs d')
        if withtheta:
            modes.append('delta/sigma vs theta')
        modes.append('normprobold')
        modes.append('normprob')
        if hasattr(reflsA[0],'icalc') and hasattr(reflsB[0],'icalc'):
            modes.append('normprob2')
        modes.append('bariobs')
        if withreso:
            modes.append('barsinthl')
            modes.append('ccsinthl')
            modes.append('ccreso')
    else:
        modes.append('bariobs')
        if withreso:
            modes.append('barsinthl')
            modes.append('ccsinthl')
            modes.append('ccreso')

    newline()
    setA = set( uniqueA.keys())
    setB = set( uniqueB.keys())
    common = setA & setB
    newline(plural.nplural(common,'common reflection'))

    for name,fullset in [ ('setA',setA), ('setB',setB) ]:
        unused = fullset - common
        if len(unused) > 0:
            newline('%s in %s not used'%(plural.reflections(unused),name))
            if args.verbose > 0:
                newline('unused reflections from',name)
                np = 0
                for hkl in unused:
                    np += 1
                    nonewline(hkl)
                    if np > DEBUGMAXREFLS:
                        newline('reflection list stopped after %s reflections'%np)
                        break
            newline()
    if scale == None:
        # Determine scale factor for setB using the
        # sum of intensities of both sets of reflections.
        sumA = 0.0
        sumB = 0.0

        for hkl in common:
            ia = uniqueA[hkl].ave('iobs')
            ib = uniqueB[hkl].ave('iobs')
            if ia > 1.0 and ib > 1.0:
                sumA += ia
                sumB += ib
        if args.verbose > 0:
            newline('sumA=',sumA,' sumB=',sumB)
        scale = sumA/sumB
        newline('scale factor (based on positive intensities) for set B %f (%f)'%(scale,1.0/scale))
    else:
        newline('scale factor (commandline) for set B %f (%f)'%(scale,1.0/scale))
        
    if args.verbose > 0:
        newline()
        np = 0
        for hkl in common:
            a = uniqueA[hkl]
            b = uniqueB[hkl]
            newline('%14s %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f' % (
                hkl, a.iobs[0], scale*b.iobs[0], a.sigma[0], scale*b.sigma[0], a.ioversig[0], b.ioversig[0]))
            np+=1
            if np > DEBUGMAXREFLS:
                newline('reflection list stopped at reflection %d'%np)
                break

    # calculate R
    for item in [ 'iobs','sigma' ]:
        teller = 0.0
        noemer = 0.0
        pairs = getpairs(item)

        for al,bl in getpairs(item):
            for a in al:
                for b in bl:
                    teller += + abs(a-scale*b)
                    noemer += + 0.5*(a+scale*b)
        newline('R(%s)=%.3f'%(item,teller/noemer))

    # create a list with all common reflection indices
    hkls = [ uniqueA[hkl].hkl for hkl in common ]

    # construct the various plots
    plotdata=[]

    if args.verbose > 0:
        newline('extra')
        newline(modes)
        newline('generating plots')
    for m in modes:
        if args.verbose > 0:
            newline(m)
        pf = PLOTrecord()
        if m == 'iobs':
            # iobsB versus iobsA
            x,y,hklp = getpoints('iobs',scale=scale)
            pf.scatter(x,y,hklp,name=m)
        elif m == 'logiobs':
            # log(iobsB) versus log(iobsA)
            x,y,hklp = getpoints('iobs',scale=scale,func=numpy.log10)
            pf.scatter(x,y,hklp,name=m)
        elif m == 'sqrtiobs':
            # sqrt(iobsB) versus sqrt(iobsA)
            x,y,hklp = getpoints('iobs',scale=scale,func=numpy.sqrt)
            pf.scatter(x,y,hklp,name=m)
        elif m == 'sigma':
            # sigmaB versus sigmaA
            x,y,hklp = getpoints('sigma',scale=scale)
            pf.scatter(x,y,hklp,name=m)
        elif m == 'logsigma':
            # log(sigmaB) versus log(sigmaA)
            x,y,hklp = getpoints('sigma',scale=scale,func=numpy.log10)
            pf.scatter(x,y,hklp,name=m)
        elif m == 'iobs/sigma' :
            # iobsB/sigmaB versus iobsA/sigmaA
            x,y,hklp = getpoints('ioversig')
            pf.scatter(x,y,hklp,name=m)
        elif m.startswith('delta/sigma'):
            IA = getlistarray(uniqueA,'iobs',common)
            IB = scale * getlistarray(uniqueB,'iobs',common)
            SA = getlistarray(uniqueA,'sigma',common)
            SB = scale * getlistarray(uniqueB,'sigma',common)
            delta = IA-IB
            sigma = numpy.sqrt( SA**2+SB**2 )
            delsig = delta/sigma
            y = delsig
            if m.endswith(' logI'):
                x,hklp = getsingles('iobs',func=numpy.log10)
            elif m.endswith(' sinthl'):
                x,hklp = getsingles('sinthl')
            elif m.endswith(' d'):
                x,hklp = getsingles('d')
            elif m.endswith(' theta'):
                x,hklp = getsingles('theta')
                x = x/DEGREES
            pf.scatter(x,y,name=m,samescale=False)
        elif m == 'normprobold' :
            IA = getlistarray(uniqueA,'iobs',common)
            IB = scale * getlistarray(uniqueB,'iobs',common)
            SA = getlistarray(uniqueA,'sigma',common)
            SB = scale * getlistarray(uniqueB,'sigma',common)
            delta = IA-IB
            sigma = numpy.sqrt( SA**2+SB**2 )
            delsig = delta/sigma
            x = NormData(delsig)
            y = delsig
            pf.scatter(x,y,name=m)
        elif m == 'normprob' :
            IA = getlistarray(uniqueA,'iobs',common)
            IB = scale * getlistarray(uniqueB,'iobs',common)
            SA = getlistarray(uniqueA,'sigma',common)
            SB = scale * getlistarray(uniqueB,'sigma',common)
            sigma = numpy.sqrt( SA**2+SB**2 )
            x,y,label = normprob(IA,IB,sigma)
            pf.scatter(x,y,name=m,label=label)
        elif m == 'normprob2':
            hor1,ver1,label1 = normprob(
                getlistarray(uniqueA,'iobs',setA),
                getlistarray(uniqueA,'icalc',setA),
                getlistarray(uniqueA,'sigma',setA)
                )
            label1 = 'a: '+label1
            hor2,ver2,label2 = normprob(
                getlistarray(uniqueB,'iobs',setB),
                getlistarray(uniqueB,'icalc',setB),
                getlistarray(uniqueB,'sigma',setB)
                )
            label2 = 'b: '+label2
            pf.scatter2(hor1,ver1,label1,hor2,ver2,label2,name=m)
        elif m == 'bariobs':
            hor = getlistarray(uniqueA,'iobs',common)
            bins,inds = GetBins(args.nbin,hor)
            # bins contains the bin upper limits 
            # inds -> bin number
            # newline(m,bins)
            if args.verbose > 0:
                newline(m,'upper boundaries for bins:')
                newline(' IntensityA',bins)
            # count (Sum(1.0) per bin)
            bc1 = numpy.bincount(inds)
            if args.verbose > 0:
                newline(' nr of reflections per bin',bc1)
            values = getlistarray(uniqueA,'iobs',common)
            # calculate sum(IobsA)  per bin
            bc2 = numpy.bincount(inds,values)
            if args.verbose > 1:
                newline('bc2',bc2)
            values = scale * getlistarray(uniqueB,'iobs',common)
            # calculate sum(IobsB) per bin
            bc3 = numpy.bincount(inds,values)
            # calculate sum(iobsA)/sum(iobsB)
            bc4 = numpy.nan_to_num(bc2/bc3)
            # no need to normalize with bc1 (both numerator and denominator)
            pf.bar(bins, bc4,name='IA/IB vs. IA bins')
        elif m == 'barsinthl':
            hor = getlistarray(uniqueA,'sinthl',common)
            bins,inds = GetBins(args.nbin,hor)
            if args.verbose > 0:
                newline(m,'upper boundaries for bins:')
                newline(' sinthl',bins)
                newline(' reso',1.0/(2.0*bins))
                if args.wavelength:
                    newline(' theta',numpy.arcsin(bins*args.wavelength)/DEGREES)
            bc1 = numpy.bincount(inds)
            if args.verbose > 0:
                newline(' nr of reflections per bin',bc1)
            values = getlistarray(uniqueA,'iobs',common)
            # calculate sum(IobsA) per bin
            bc2 = numpy.bincount(inds,values)
            values = scale * getlistarray(uniqueB,'iobs',common)
            # calculate sum(IobsB) per bin
            bc3 = numpy.bincount(inds,values)
            # calculate sum(iobsA)/sum(iobsB)
            bc4 = numpy.nan_to_num(bc2/bc3)
            # no need to normalize with bc1 (both numerator and denominator)
            if args.verbose > 0:
                newline('bc4barsinthl',bc4)
            pf.bar(bins, bc4,name=m)
        elif m == 'ccsinthl':
            hor = getlistarray(uniqueA,'sinthl',common)
            bins,inds = GetBins(args.nbin,hor)
            if args.verbose > 0:
                newline(m,'upper boundaries for bins:')
                newline(' sinthl',bins)
                newline(' reso',1.0/(2.0*bins))
                if args.wavelength:
                    newline(' theta',numpy.arcsin(bins*args.wavelength)/DEGREES)
            bc1 = numpy.bincount(inds)
            if args.verbose > 0:
                newline(' nr of reflections per bin',bc1)
            valuesa = getlistarray(uniqueA,'iobs',common)
            valuesb = scale * getlistarray(uniqueB,'iobs',common)
            bc4 = []
            for ibin in range(args.nbin):
                va = [ valuesa[i] for i in range(len(inds)) if inds[i] == ibin ]
                vb = [ valuesb[i] for i in range(len(inds)) if inds[i] == ibin ]
                bc4.append( numpy.corrcoef(va,vb)[0,1] )
            if args.verbose > 0:
                newline('bc4',bc4)
            bc4 = numpy.array(bc4)
            pf.bar(bins, bc4,name=m)
        elif m == 'ccreso':
            hor = getlistarray(uniqueA,'d',common)
            bins,inds = GetBins(args.nbin,hor)
            if args.verbose > 0:
                newline(m,'upper boundaries for bins:')
                newline(' reso',bins)
            bc1 = numpy.bincount(inds)
            if args.verbose > 0:
                newline(' nr of reflections per bin',bc1)
            valuesa = getlistarray(uniqueA,'iobs',common)
            valuesb = scale * getlistarray(uniqueB,'iobs',common)
            bc4 = []
            for ibin in range(args.nbin):
                va = [ valuesa[i] for i in range(len(inds)) if inds[i] == ibin ]
                vb = [ valuesb[i] for i in range(len(inds)) if inds[i] == ibin ]
                bc4.append( numpy.corrcoef(va,vb)[0,1] )
            if args.verbose > 1:
                newline('bc4',bc4)
            bc4 = numpy.array(bc4)
            pf.bar(bins, bc4,name=m)

        plotdata.append(pf)

    if args.verbose > 0:
        newline('draw')
    pagesize = (args.pagehor/INCH,args.pagever/INCH)
    fontsize=12.0
    args.charheight = fontsize/72.0/12
    args.charheight = 1.5*args.charheight
    aspect = args.pagehor/args.pagever
    if args.verbose > 1:
        newline('backend=%s'%plt.get_backend())

    if args.verbose > 1:
        newline('before default=regular')
        for key in sorted(plt.rcParams):
            if key.startswith('math'):
                newline(key,plt.rcParams[key])
    rc('mathtext', default='regular')
    if 'patch.force_edgecolor' in mpl.rcParams:
        mpl.rcParams['patch.force_edgecolor'] = True

    if args.verbose > 1:
        newline('after default=regular')
        for key in sorted(plt.rcParams):
            if key.startswith('math'):
                newline(key,plt.rcParams[key])

    fig=plt.figure( figsize=pagesize )
    fig.canvas.manager.set_window_title('loglog')

    almodes = []
    for m in modes:
        almodes.append(alias(m))
    alinfo = []
    for m in info:
        alinfo.append(alias(m))

    margin = 0.01
    left = 0.28
    right = 1.0-0.05-margin
    bottom = 0.09
    top = 1.0-0.03-margin*aspect
    plt.subplots_adjust(left=left, right=right, bottom=bottom, top=top)
    ax = fig.add_subplot(111)
    ndxp = -1
    ndxh = (-1,0,0)

    labelit(ax)
#    drawhkl()

    #  points are pickable
    cid = fig.canvas.mpl_connect('pick_event', on_pick)
    fig.canvas.mpl_connect('key_press_event', pressrelease)
    #fig.canvas.mpl_connect('key_release_event', pressrelease)

    # clickables from bottom up
    bottom,buttons = createButtons(0.01)
    bottom,radios = createRadios(bottom)

    # legenda's top down
    
    bottom = 0.99
    key = list(uniqueA.keys())[0]
    # newline(uniqueA[key])
    a = [ (item[0],'') for item in uniqueA[key].shortlegenda() ]
    bottom,legendaMain = CreateTable(bottom,a)

    ainfo = info[:]
    if hasattr(reflsA[0],'icalc'):
        ainfo.append('icalc')
    a = [ (item,'') for item in ainfo ]
    bottom,legendaOne = CreateTable(bottom,a,title='a:'+os.path.basename(filelist[0]))

    binfo = info[:]
    if hasattr(reflsB[0],'icalc'):
        binfo.append('icalc')
    b = [ (item,'') for item in binfo ]
    bottom,legendaTwo = CreateTable(bottom,b,title='b:'+os.path.basename(filelist[1]))

    triggerSwitch(almodes[mode])
    plt.show()

if __name__ == '__main__':
    main()
