"""
Event logger
"""
#
__version__ = '$Id: log.py,v 1.8 2020/09/02 10:43:35 schreurs Exp $'
#
# (C) 1998--2004 Rob W. W. Hooft, Bruker AXS BV
#
import sys,time,re,os

from misc import config

# List of severities
SEV_EMERGENCY=0
SEV_ALERT=1
SEV_BUG=2
SEV_CRIT=3
SEV_ERROR=4
SEV_WARNING=5
SEV_NOTICE=6
SEV_INFO=7
SEV_DEBUG=8

severstr={0:"EMERGENCY",
          1:"ALERT",
          2:"PROBABLE PROGRAM BUG",
          3:"FATAL ERROR",
          4:"ERROR",
          5:"WARNING",
          6:"NOTICE",
          7:"INFORMATIONAL",
          8:"DEBUGGING"
         }

# Name of the program
_prog=sys.argv[0]
if len(_prog):
    if _prog[0]==os.sep:
        _prog=os.path.basename(_prog)
    if _prog[-3:]==".py":
        _prog=_prog[:-3]
    if _prog[-4:]==".pyc":
        _prog=_prog[:-4]
    if _prog=="top":
        _prog="collect"
    if _prog=="dtop":
        _prog="ndisp"

class BugTrap(Exception):
    """Exception that can be raised by the BUG function"""
    pass

class Log:
    """
    Class used to access the log file, and write events there.

    This is used in this module, and should not need to be instantiated
    outside.
    """
    def __init__(self,filename,level=6):
        """
        Constructor.

        parameters:

          **filename** -- the filename of the log file.

          **level** -- the highest severity level that will be logged.
        """
        self.filename=filename
        self.level=level
        self.modulelevel={}
        # A lookup table of all 'write errors' that we did already show.
        self.wrerror={}
        self.last=''
        
    def log(self,module,severity,message):
        """
        Log a message.

        parameters:

          **module** -- the name of the module in the program where the event
                        occurred.

          **severity** -- the severity of the event, between 0 and 8 with higher
                          numbers being less fatal. There are symbolical names
                          SEV_* for these levels available in this module.

          **message** -- the text describing the event.
        """
        if not self.filename:
            return
        if module=="__main__" or module=="projtls" or module is None:
            module=""
        else:
            module="."+module
        #if self.modulelevel.has_key(module):
        if module in self.modulelevel:
            if severity>self.modulelevel[module]:
                return
        else:
            if severity>self.level:
                return
        modsev=_prog+module+": "+"  "+severstr[severity]
        if self.last==modsev:
            txt=""
        else:
            import projtls
            txt="\n--- "+time.ctime(time.time())+" "+projtls.host()+" --- "+modsev+"\n"
        # Put 8 spaces in front of every message line.
        message=re.sub("\n","\n        ",message)
        txt=txt+"      * "+message+"\n"
        try:
            f=open(self.filename,'a')
            f.write(txt)
            self.last=modsev
            f.close()
        except(IOError) as arg:
            s=str(arg)
            # Report errors only once.
            if not s in self.wrerror:
                self.wrerror[s]=None
                print("Cannot write to '%s': %s"%(self.filename,s))

_globlog=Log(filename=config.eventlogfile)
log=_globlog.log

def ERROR(s,exit=None,severity=SEV_ERROR,name=None,terminal=1):
    """
    Register an ERROR event (severity 4).

    parameters:

      **s** -- the string of the error message.

      **exit** -- if not 'None', this routine does not return, but exits with the
                  given exit status.
                  
      **severity** -- do not touch.

      **name** -- the name of the module where the error occurred

      **terminal** -- if 1, the error is logged to the terminal too.
    """
    log(name,severity,s)
    if terminal:
        ss=s.split('\n')
        if severity>SEV_WARNING:
            file=sys.stdout
        else:
            file=sys.stderr
        prefix=severstr[severity]+": "
        file.write(prefix+ss[0]+'\n')
        for line in ss[1:]:
            file.write(' '*len(prefix)+line+'\n')
    if exit is not None:
        if severity<=SEV_ERROR:
            from gui import setup
            #from gui import command
            if setup.root():
                command.errorbox('root',s)
        sys.exit(exit)

def WARNING(s,exit=None,severity=SEV_WARNING,name=None,terminal=1):
    """
    Register a WARNING event (severity 5). For parameters see ERROR
    """
    ERROR(s,exit,severity,name,terminal)

def FATAL(s,exit=1,severity=SEV_CRIT,name=None,terminal=1):
    """
    Register a FATAL event (severity 3). For parameters see ERROR
    """
    ERROR(s,exit,severity,name,terminal)

def BUG(s,exit=None,severity=SEV_BUG,name=None,terminal=1,crash=1):
    """
    Register a BUG event (severity 2). For parameters see ERROR

    if crash==1, and exit is None, the program will generate a
    stack trace.
    """
    ERROR(s,exit,severity,name,terminal)
    if crash:
        raise BugTrap(s)

def NOTE(s,exit=None,severity=SEV_NOTICE,name=None,terminal=1):
    """
    Register a NOTE event (severity 6). For parameters see ERROR
    """
    ERROR(s,exit,severity,name,terminal)

def INFO(s,exit=None,severity=SEV_INFO,name=None,terminal=1):
    """
    Register an INFO event (severity 7). For parameters see ERROR
    """
    ERROR(s,exit,severity,name,terminal)

def UNKARG(argname,extratext=None):
    """
    Special case of WARNING, to be used for unknown command line arguments.
    """
    if extratext:
        WARNING("Unknown argument '%s' ignored.\n%s"%(argname,extratext))
    else:
        WARNING("Unknown argument '%s' ignored."%argname)

if __name__=="__main__":
    INFO("module loaded",name=__name__)
