"""
Some matrix/vector tools.

 The collect python library uses the numpy python extension for all of its
 major array handling. Sometimes the standard functions are a bit awkward to use
 or in a difficult location for simple 3x3-matrix and 3-vector work.  The
 "matrix" module assembles some matrix/vector routines.
"""
#
__version__='$Id: matrix.py,v 1.6 2020/08/30 10:55:16 schreurs Exp $'
#
# (C) 1999--2004 Rob W.W. Hooft, Bruker AXS BV
#
#import Numeric,LinearAlgebra
#import numpy.oldnumeric as Numeric
#import numpy.linalg
import numpy

def inverse(m):
    """
    Calculate the matrix inverse.
    """
    pass
inverse=numpy.linalg.inv

def determinant(m):
    """
    Calculate the matrix determinant.
    """
    pass
determinant= numpy.linalg.det

def transpose(m):
    """
    Calculate the transposed matrix.
    """
transpose=numpy.transpose

def matmat(m1,m2):
    """
    Matrix-multiplication of 2 simple matrices
    """
multiply=numpy.dot

matmat=numpy.dot

##  def matmat3(m1,m2):
##      m=numpy.zeros((3,3),m1.typecode())
##      r3=range(3)
##      for i in r3:
##          for k in r3:
##              m[i,k]=numpy.add.reduce(m1[i,:]*m2[:,k])
##      return m

def volume(m):
    """Determinant for 3x3 matrix"""
    v1=m[0]
    v2=m[1]
    v3=m[2]
    return ((v1[1]*v2[2]-v1[2]*v2[1])*v3[0]
            +(v1[2]*v2[0]-v1[0]*v2[2])*v3[1]
            +(v1[0]*v2[1]-v1[1]*v2[0])*v3[2])
    
def matvec(m,v):
    """Multiply a matrix with a 3D vector, or any array of vectors"""
    v=numpy.array(v,copy=0)
    ax=list(range(1,len(v.shape)))+[0]
    return numpy.transpose(numpy.inner(m,v),axes=ax)

def vectorlength(v):
    """Determine length of single 3D vector or each out of an array of vectors"""
    if v.shape[-1]!=3:
        raise ValueError("Invalid shape for v in vectorlength")
    return numpy.sqrt(v[...,0]*v[...,0]+v[...,1]*v[...,1]+v[...,2]*v[...,2])

def vectorangle(v1,v2):
    """Determing the angle between two 3D vectors (or arrays of vectors)."""
    if v1.shape[-1]!=3:
        raise ValueError("Invalid shape for v1 in vectorangle")
    if v1.shape!=v2.shape and v1.shape!=(3,) and v2.shape!=(3,):
        raise ValueError("Unequal shapes for v1 and v2 in vectorangle")
    cross0=v1[...,1]*v2[...,2]-v1[...,2]*v2[...,1]
    cross1=v1[...,2]*v2[...,0]-v1[...,0]*v2[...,2]
    cross2=v1[...,0]*v2[...,1]-v1[...,1]*v2[...,0]
    cross=numpy.sqrt(cross0*cross0+cross1*cross1+cross2*cross2)
    del cross0,cross1,cross2
    dot=v1[...,0]*v2[...,0]+v1[...,1]*v2[...,1]+v1[...,2]*v2[...,2]
    return numpy.arctan2(cross,dot)

def dotproduct(v1,v2):
    return v1[...,0]*v2[...,0]+v1[...,1]*v2[...,1]+v1[...,2]*v2[...,2]

def vectorproduct(v1,v2):
    """Calculate cross product between 2 3D vectors (or arrays of vectors)."""
    cross0=v1[...,1]*v2[...,2]-v1[...,2]*v2[...,1]
    cross1=v1[...,2]*v2[...,0]-v1[...,0]*v2[...,2]
    cross2=v1[...,0]*v2[...,1]-v1[...,1]*v2[...,0]
    return numpy.array([cross0,cross1,cross2])

def directionangle(v1,v2):
    """Calculate the angle between two directions (0-90 degrees)"""
    d=vectorangle(v1,v2)
    if d<numpy.pi/2:
        return d
    else:
        return numpy.pi-d

def Xrot(a,typecode='d'):
    """Construct a rotation matrix rotating 'a' around the X axis"""
    sa=numpy.sin(a)
    ca=numpy.cos(a)
    return numpy.array([[1.0,0,0],[0,ca,sa],[0,-sa,ca]],typecode)

def Yrot(a,typecode='d'):
    """Construct a rotation matrix rotating 'a' around the Y axis"""
    sa=numpy.sin(a)
    ca=numpy.cos(a)
    return numpy.array([[ca,0,-sa],[0,1.0,0],[sa,0,ca]],typecode)

def Zrot(a,typecode='d'):
    """Construct a rotation matrix rotating 'a' around the Z axis"""
    sa=numpy.sin(a)
    ca=numpy.cos(a)
    return numpy.array([[ca,sa,0],[-sa,ca,0],[0,0,1.0]],typecode)

def Vrot(v,a,typecode='d'):
    """Rotate 'a' around a given vector 'v'"""
    v=numpy.array(v)
    v=v/vectorlength(v)
    rotmat=numpy.zeros((3,3),typecode)
    ca=numpy.cos(a)
    ca1=1.0-ca
    sa=numpy.sin(a)
    for i in range(3):
        rotmat[i,i] = v[i]**2 * ca1 + ca
        j = (i+1)%3
        k = (j+1)%3
        f = v[j]*v[k]*ca1
        g = v[i]*sa
        rotmat[j,k] = f-g
        rotmat[k,j] = f+g
    return rotmat

def samematrix(m1,m2,crit=0.05):
    """Compare matrices, return 1 if they are approximately identical."""
    return numpy.maximum.reduce(numpy.absolute(numpy.ravel(m1-m2)))<crit

def integermatrix(m,crit=0.1):
    """Return 1 if m is approximately consisting of integer coefficients."""
    v=numpy.ravel(m)
    v=numpy.absolute(v)%1.0
    for i in range(len(v)):
        if v[i]>0.5:
            v[i]=v[i]-1.0
    x=numpy.maximum.reduce(numpy.absolute(v))
    return x<crit

def dotprod(v1,v2):
    """Calculate scalar product of 2 3D vectors."""
    return numpy.add.reduce(v1*v2)

def crossprod(v1,v2):
    """Calculate vector product of 2 3D vectors."""
    return numpy.array([v1[1]*v2[2]-v1[2]*v2[1],
                          v1[2]*v2[0]-v1[0]*v2[2],
                          v1[0]*v2[1]-v1[1]*v2[0]])
