#
from __future__ import print_function
"""Some basic tools for the 'evpy' package

This module contains a single "constant" value that is used for
all angle handling in the program: 'projtls.degrees'. This is
the conversion factor to convert from degrees to radians and back. Please
note that ALL angles in the program are supposed to be stored in radians,
with exception of those in the configuration files, and thus those in the
"misc.config" module.

Examples::

  import projtls
  theta=60*projtls.degrees
  print("Theta= %.4f degrees"%theta/projtls.degrees)

To remember the convention, "*projtls.degrees" can be pronounced as "to
degrees", and "/projtls.degrees" as "in degrees".
"""
#
__version__='$Id: projtls.py,v 1.17 2021/09/14 14:36:40 schreurs Exp $'
#
#
# (C) 1997--2004 Rob W.W. Hooft, Bruker AXS BV
#
import math,os,sys

if sys.version[:1]<'2':
    print("You are running a version of the base software that is too old")
    print("Please upgrade to the latest base (at least release 5) before")
    print("running this version of the collect programs")
    sys.exit(1)

if sys.version_info.major >= 3:
    long=int
degrees=math.pi/180.0
radians=1.0

root=os.getenv('EVPYROOT')
if not root:
    root=sys.path[1]
#print("root=",root)

class ScanTextError(Exception):
    """Exception raised when ReScanText encounters a parsing problem.
    """
    pass

def ReScanText(intext,what,into=None,allowunknown=0,catchAll=None):
    """Scan some textlines for some regular expressions, storing some variables in an object

    once a line has matched,
    no more regular expressions are tested.

    - parameters:

        **intext** --   list of text-lines of input, each to be scanned
                 (if a single string is given, it is split into lines first)

        **what** --     data structure describing what to search

    - optional parameters:
    
        **into** --     any object to store the attributes that were found.
                   If it is not specified, an empty Record() is created.

        **allowunknown** -- if set to 1, will just complain a bit if a line
                            is not understood. If set to 2, will not even
                            complain. If set to 0, will raise an exception.

    - return value:
        into

    side effect: into is changed.

    The "what" structure is composed of:

      **what** = tuple of trials
      
      **trial** = tuple of ("re" regular expression object, tuple of groups)
      
      **group** = tuple of (group number, type, attribute name)
      
      **type** = "f" (float) | "l" (logical [Yes/No]) | "i" (integer) |
                "s" (string) | "a" (angle in degrees) | "t" (list of strings)
    """
    if into is None:
        into=Record()
    if type(intext)==type(''):
        intext=intext.split('\n')
    lasti=-1
    for line in intext:
        match=None
        if lasti >= 0:
            # Try the same as last time
            if lasti<len(what):
                expr=what[lasti]
                match=expr[0].match(line)
                # if match:
                #     print("Same")

            # Try the next.
            if not match and lasti<len(what)-1:
                expr=what[lasti+1]
                match=expr[0].match(line)
                if match:
                    lasti+=1
                    # print("Next")
        if not match:
            #print("Search",line)
            lasti=0
            for expr in what:
                match=expr[0].match(line)
                if match:
                    break
                lasti+=1
        if not match:
            if catchAll:
                if hasattr(into,catchAll):
                    k=getattr(into,catchAll)+[line]
                else:
                    k=[line]
                setattr(into,catchAll,k)
                continue

            if allowunknown:
                if allowunknown==1:
                    print("Not understood",line)
            else:
                raise ScanTextError("Not understood",line)
            continue
        setattr(into,'regex',lasti)
        for var in expr[1]:
            try:
                try:
                    k=match.group(var[0])
                except IndexError:
                    print(line)
                    print("No such group %d"%var[0])
                    raise
                except TypeError:
                    print(var)
                    print(line)
                    print("var has wrong type?")
                    raise
                if var[1]=='f':
                    k=float(k)
                elif var[1]=='a':
                    k=float(k)*degrees
                elif var[1]=='i':
                    k=int(k)
                elif var[1] == 'L':
                    k = long(k)
                elif var[1]=='l':
                    if (k=="Yes"):
                        k=1
                    else:
                        k=0
                elif var[1]=='t': # Array of strings
                    if hasattr(into,var[2]):
                        k=getattr(into,var[2])+[k]
                    else:
                        k=[k]
                setattr(into,var[2],k)
            except ValueError:
                print("WARNING: Scan error '%s'"%line)
    return into

class Record:
    """Empty class for misc use"""
    _format={}
    
    def __init__(self,**kw):
        for key,value in kw.items():
            setattr(self,key,value)

    def __repr__(self):
        t=[]
        for i in dir(self):
            if i[0]!="_" and i != "copy":
                if self._format.has_key(i):
                    if self._format[i]:
                        t.append("%s=%s"%(i,self._format[i]%getattr(self,i)))
                else:
                    t.append("%s=%s"%(i,repr(getattr(self,i))))
        return "Record: "+' '.join(t)

    def copy(self):
        import copy
        return copy.copy(self)
    
class Cache:
    """General Data cache"""
    def __init__(self,keep=5):
        """Constructor.

        **keep** -- Number of results to cache
        """
        self.cache=[]
        self.keep=keep

    def __call__(self,operation,input):
        """Call of cache object.

           - parameters:
           
               **operation** -- the real function to call if the result is
                                not yet known. This should be a one-argument
                                stateless function.
                                
               **input** -- the result of operation(input) is requested.

           - Return value is the value of operation(input). If the result
             was in the cache, the routine operation is not called to get
             this result.
        """
        for cin,cout in self.cache:
            if cin==input:
                return cout
        output=operation(input)
        self.cache[0:0]=[(input,output)]
        self.cache=self.cache[0:self.keep]
        return output

    def delete(self,input):
        """Delete cache item.

           - parameters:
           
               **input** -- operation(input) should be removed from cache

           - No return value
        """
        for i in range(len(self.cache)):
            if self.cache[i][0]==input:
                del self.cache[i]
                break

def host():
    name = os.uname()[1]
    #            import socket
    #            host = socket.gethostname()
    if not name:
        name = ''
    return name

def getlogin():
    import pwd
    data = pwd.getpwuid(os.getuid())
    return data.pw_name
#    return os.getenv('LOGNAME')

def createdby(f,label,comment='#'):
    import time
    f.write("! Created by %s at %s\n"%(label,time.ctime(time.time())))
    f.write("! Host %s User %s WD %s\n"%(host(),getlogin(),os.getcwd()))

def moveoutofway(filename,dirok=0):
    """Rename the given file to a safe backup name.

    For the destination name, use the "emacs" .~#~ extension.

    - parameters:

      **filename**: the file to be renamed. If it does not exist, nothing
                    will happen.

      **dirok**: normally only plain files will be renamed away, but if
                 this option is set to 1, directories will be considered as
                 well.
    """
    if not os.path.exists(filename):
        return
    if not os.path.isfile(filename):
        if not dirok or not os.path.isdir(filename):
            return
    i=0
    while 1:
        i=i+1
        renamename=filename+".~%d~"%i
        if not os.path.exists(renamename):
            try:
                os.rename(filename,renamename)
                break
            except os.error:
                print("Could not move existing file %s out of the way"%filename)
                return
            else:
                print("Existing file %s renamed to %s"%(filename,renamename))

def myopen(filename,mode):
    """Open a file. Based on the filename extension, open it through different
       compression/decompression program pipes.

       parameters:
           filename : name of file to open.
           mode : 'r' or 'w' for read-only or write-only open.
    """
    from interface import compress
    if mode=='r':
        for ext,prog in compress.table.items():
            if filename[-len(ext):]==ext:
                return os.popen(prog+' -dc '+filename+' 2> /dev/null' ,'r')
        else:
            return open(filename,'r')
    elif mode=='w':
        moveoutofway(filename)
        for ext,prog in compress.table.items():
            if filename[-len(ext):]==ext:
                return os.popen(prog+' > '+filename,'w',0)
        else:
            return open(filename,'w')
    raise IOError("unknown filemode?")

def filesize(filename):
    """Returns file size in bytes. To get real size of compressed file, use
       the routine with the same name from the interface.compress module"""
    return os.path.getsize(filename)

def fileage(filename):
    """Get the relative file "age" (tests reverse from the "file time")""" 
    return -os.stat(filename)[8]

def filetime(filename):
    """Get the time the file was made""" 
    return os.stat(filename)[8]

def pcbyteorder():
    """Determine the byte order of the machine, to know e.g. whether KappaCCD
       images should be byte-swapped before presentation (this is needed
       when this routine returns 0)"""
    import socket
    return socket.htons(256)==1

class LongTaskTracer:
    """Allow a Tkinter GUI to interrupt a long-lasting computation.
    
       There is one instance-object of this class created by the projtls
       module, it should not be necessary to make more."""
    def __init__(self,interval=0.1):
        """Constructor.

        **interval** -- the minimum interval (in seconds) between
                        GUI update calls.
        """
        self.cancelbusy=0
        self.idletask=None
        self.lasttime=0
        self.interval=interval

    def cancelbusytask(self,but='Abort'):
        """Abort a long-lasting computation from a GUI button.

        (For the singleton instance this is available as
        "projtls.cancelbusytask")
        """
        if but in ['Abort','Cancel']:
            #print("ltt: Cancel raised!")
            self.cancelbusy=1

    def reset(self):
        """Reset the cancel request once it has been handled.
        """
        #print("ltt: reset cancel")
        self.cancelbusy=0

    def checkifcanceled(self):
        """Check whether an abort was requested from a GUI button.

        Call this routine regularly at safe points in a long calculation (For
        the singleton instance this is available as "projtls.checkifcanceled")

        No value is returned. If the long calculation should be interrupted,
        a gui.command.GuiCancel exception (subclass of KeyboardInterrupt)
        is raised.
        """
        if self.cancelbusy:
            #print("ltt: oops, and auto-reset!")
            self.cancelbusy=0
            from gui import command
            raise command.GuiCancel("Busy task canceled")
        if self.idletask:
            import time
            now=time.time()
            if now-self.lasttime>self.interval:
                self.lasttime=now
                self.idletask()

    def setidletask(self,idletask):
        """Sets the idle task that should be used to try and keep the
        GUI active. For the singleton instance, this is automatically called
        by gui.setup.setup()"""
        self.idletask=idletask
        self.lasttime=0

# The singleton instance.
_ltt=LongTaskTracer()
cancelbusytask=_ltt.cancelbusytask
checkifcanceled=_ltt.checkifcanceled
resetcancel=_ltt.reset

def freespace(path):
    """ freespace(path) -> integer
    Return the number of bytes available to the user on the file system
    pointed to by path."""
    s = os.statvfs(path)
    return s.f_bavail * long(s.f_bsize)

def stringtofloats(s):
    """Reads space-separated floating point values from a string.
    Returns as many floats as there are. Return value is a tuple
    of the values"""
    return tuple(map(float,s.split()))

def timestamp():
    """Return a string in the form of 'Sep14 12:15:46.985: ' representing
    the current time. This is meant for logging (network) events."""
    import time
    t=time.time()
    lt=time.localtime(t)
    return time.strftime("%b%d %H:%M:%S",lt)+("%.3f: "%(t-int(t)))[1:]

def timestamps():
    """Return a string in the form of 'Sep14 12:15:46' representing
    the current time. This is meant for less accurate logging."""
    import time
    t=time.time()
    lt=time.localtime(t)
    return time.strftime("%b%d %H:%M:%S",lt)

class CalledElement:
    """Help to delay list evaluation to the latest possible moment.

    This class is useful if you have a function returning a list, and
    you might need a certain element from that list later. Encapsulating
    the call in a CalledElement will make sure that the function returning
    the list is only called when a reference is being made to this element.
    """
    def __init__(self,listfunc,itemno):
        """Constructor.
        
        **listfunc** -- A function returning a list

        **itemno** -- the element number in the list we are interested in
        """
        self.listfunc=listfunc
        self.itemno=itemno
        self.result=None
        self.filling=0

    def _fill(self):
        self.filling=1
        if self.result is None:
            self.result=self.listfunc()[self.itemno]
        self.filling=0

    def __call__(self,*arg,**kw):
        self._fill()
        return apply(self.result,arg,kw)

    def __getattr__(self,attr):
        if attr=='filling':
            return 0
        if self.filling:
            raise AttributeError("Filling %s %s",self,attr)
        self._fill()
        return getattr(self.result,attr)

    def __repr__(self):
        self._fill()
        if self.result.__doc__:
            return self.result.__doc__.split('\n')[0]
        else:
            return repr(self.result)

def youngest(pattern=None,files=None,exclude=()):
    """Return the youngest file from a set.

       Can be called with:
       
          **string in pattern** -- return the youngest file that matches the pattern

          **list or tuple of strings in pattern** -- return the youngest file that
                         matches any of the patterns
                         
          **no arguments** -- return the youngest file in the current directory
          
          **a string in pattern and a list of filenames in files** --
             return the youngest file from the list of files that matches
             the pattern
                  
          **a list of filenames in files** --
             return the youngest file from the list

       There is a function of the same name in the misc.compress module
       that will also find any compressed files.
       """
    import fnmatch
    if files is None:
        if pattern is None:
            pattern="*"
        import glob
        if type(pattern)==type([]) or type(pattern)==type(()):
            files=[]
            for pat in pattern:
                files.extend(glob.glob(pat))
        else:
            files=glob.glob(pattern)
    elif pattern:
        nfiles=[]
        for file in files:
            if fnmatch.fnmatch(file,pattern):
                nfiles.append(file)
        files=nfiles
    if files:
        youngestfile=None
        yage=None
        for fn in files:
            try:
                if yage is None or fileage(fn)<yage:
                    for efn in exclude:
                        if fnmatch.fnmatch(fn,efn):
                            break
                    else:
                        yage=fileage(fn)
                        youngestfile=fn
            except IOError: # E.g. symbolic link to non-existing file
                pass
        return youngestfile
    else:
        return None

def oldest(pattern=None,files=None,exclude=()):
    """Return the oldest file from a set.

       Can be called with:
       
          **string in pattern** -- return the oldest file that matches the pattern

          **list or tuple of strings in pattern** -- return the oldest file that
                         matches any of the patterns
                         
          **no arguments** -- return the oldest file in the current directory
          
          **a string in pattern and a list of filenames in files** --
             return the oldest file from the list of files that matches
             the pattern
                  
          **a list of filenames in files** --
             return the oldest file from the list

       There is a function of the same name in the misc.compress module
       that will also find any compressed files.
       """
    import fnmatch
    if files is None:
        if pattern is None:
            pattern="*"
        import glob
        if type(pattern)==type([]) or type(pattern)==type(()):
            files=[]
            for pat in pattern:
                files.extend(glob.glob(pat))
        else:
            files=glob.glob(pattern)
    elif pattern:
        nfiles=[]
        for file in files:
            if fnmatch.fnmatch(file,pattern):
                nfiles.append(file)
        files=nfiles
    if files:
        oldestfile=None
        yage=None
        for fn in files:
            try:
                if yage is None or fileage(fn)>yage:
                    for efn in exclude:
                        if fnmatch.fnmatch(fn,efn):
                            break
                    else:
                        yage=fileage(fn)
                        oldestfile=fn
            except IOError: # E.g. symbolic link to non-existing file
                pass
        return oldestfile
    else:
        return None

def wordsfromfile(filename):
    """Read all non-comment lines from the given file.

       Return value: list of words"""
    f=open(filename,"r")
    lines=filter(lambda l:l[0]!='!' and l[0]!='#',f.readlines())
    f.close()
    lines=' '.join(lines)
    return lines.split()

class timer:
    """Poor man's profiler"""
    def __init__(self,string=None,state=1,showtime=1):
        """
        Constructor.

        parameters:

          **string** -- first message to print

          **state** -- 1=on (will display messages) 0=off (no output)
        """
        import time
        self.clock=time.clock
        self.state=state
        self.showtime=showtime
        self.first(string)

    def setstate(self,state):
        """
        Change the state.

        parameter:
        
          **state** -- 1=on (will display messages) 0=off (no output)
        """
        self.state=state
        
    def _out(self,string):
        """
        Print the string.
        """
        if self.state:
            sys.stdout.write(string)
            sys.stdout.flush()
            
    def first(self,string):
        """
        Start the first timing cycle

        parameter:

          **string** -- a short string describing what the program will start
                        doing now.
        """
        self.reference=self.clock()
        if string:
            self._out(string+"... ")

    def intermediate(self):
        """Print a sub-time"""
        self._out(self()+"... ")
        
    def next(self,string):
        """
        Complete the previous timing line, starting a new one

        parameter:

          **string** -- a short string describing what the program will start
                        doing next.
        """
        self._out(self()+"\n")
        self.first(string)

    def last(self):
        """
        Complete the last timing line
        """
        self.next(None)
        
    def __call__(self):
        """Timer tick.

        Returns a string representing the CPU time used since the last tick,
        thereby resetting the counter. For internal use by the class.

        if showtime was set to 0 by the constructor, an empty string is returned
        instead.
        """
        now=self.clock()
        s="%.2f seconds"%(now-self.reference)
        self.reference=now
        if self.showtime:
            return s
        else:
            return ""

def MappedOrString(f,size):
    """Return a memory mapped R/O part of a file, or if that doesn't work
       the same data as a string."""
    return f.read(size)

def newline(*s):
    if len(s):
        print(*s)
    else:
        print()

def nonewline(*s):
    print(*s,end=' ')

